#!/usr/bin/env python3
"""
2013.05.03 a child class of MatrixFile. used to describe Beagle likelihood file format, which looks like:
    marker alleleA alleleB 1000_709_1996093_GA_vs_524 1000_709_1996093_GA_vs_524 1000_709_1996093_GA_vs_524 1001_710_1995025_GA_vs_524 1001_710_1995025_GA_vs_524 1001_710_1995025_GA_vs_524 1002_711_2001039_GA_vs_524
    Contig791:1086 C A 1 0 0 0.9997 0.0003 0 0
    Contig791:1649 G C 1 0 0 1 0 0 0
    Contig791:4084 A C 1 0 0 1 0 0 0
    Contig791:4118 A G 1 0 0 1 0 0 0
    Contig791:4143 C A 1 0 0 1 0 0 0
    Contig791:4168 G C 1 0 0 0.9999 0.0001 0 0
    Contig791:4203 C G 1 0 0 1 0 0 0

Example:
    beagleFile = BeagleGenotypeFile(path='/tmp/input.bgl.gz')
    beagleFile.readInAllHaplotypes()
    
    for individualID, firstHaplotypeIndex in beagleFile.snpData.col_id2col_index.items():
        haplotypeList = []
        for j in range(firstHaplotypeIndex, firstHaplotypeIndex+self.ploidy):
            haplotypeList.append(beagleFile.snpData.data_matrix[:,j]) 
        # another way
        #haplotypeList = beagleFile.getHaplotypeListOfOneSample(individualID)
    
    reader = MatrixFile(path='/tmp/input.txt', mode='r')
    reader = MatrixFile('/tmp/input.txt', mode='r')
    reader.constructColName2IndexFromHeader()
    for row in reader:
        row[reader.getColName2IndexFromHeader('KID')]
    
    inf = utils.openGzipFile(path, mode='r')
    reader = MatrixFile(file_handle=inf)
    
    #2013.2.1 writing
    writer = MatrixFile('/tmp/output.txt', mode='w', delimiter='\t')
    writer.writeHeader(...)
    writer.writerow(row)
    writer.close()

"""
import sys, os
import copy, numpy
from palos import utils, PassingData
from palos.ProcessOptions import ProcessOptions
from palos.polymorphism.SNP import SNPData
from palos.io.MatrixFile import MatrixFile

class BeagleGenotypeFile(MatrixFile):
    __doc__ = __doc__
    option_default_dict = copy.deepcopy(MatrixFile.option_default_dict)
    option_default_dict.update({
        ('delimiter', 0, ): [' ', '', 1, 'delimiter for Beagle likelihood format is single-space'],\
        ('phased', 0, int): [0, '', 1, 'whether the genotypes are phased or not'],\
        ('ploidy', 0, int): [2, '', 1, 'ploidy of the organism. = #genotype-columns for one individual'],\
        })
    def __init__(self, path=None, **keywords):
        MatrixFile.__init__(self, path=path, **keywords)
        self.header = None
        self.col_name2index = None	#key is sampleID, value is index of first haplotype
        
        self.sampleIDList = []
        self.sampleID2index = {}	#same as col_name2index
        self.locusIDList = []
        self.haplotypeMatrix = []
        
        self.snpData = None	#to store everything above . SNPData type
    
    def _completeSNPData(self):
        """
        2013.06.05
        """
        if self.snpData is None:
            if self.mode=='r':
                self.readInAllHaplotypes()
            else:	#writing mode
                self._consolidateHaplotypeMatrixIntoSNPData()
    
    def getSampleIndexGivenSampleID(self, sampleID=None):
        """
        2013.06.05
        """
        return self.getColIndexGivenColHeader(colHeader=sampleID)
    
    def constructColName2IndexFromHeader(self):
        """
        2013.05.03
            First two columns are for designators.
            one diploid sample occupies two columns.
            
        I id 1466_1867_2001024_GA_vs_524copy3 1466_1867_2001024_GA_vs_524copy3 1243_1338_2005057_GA_vs_524copy3 1511_639_1987079_GA_vs_524copy3 1511_639_1987079_GA_vs_524copy3 1016_725_1995116_GA_vs_524copy7 1521_767_1993041_GA_vs_524copy4 1521_767_1993041_GA_vs_524copy4
        M Contig791:1086 C C C A A C C C
        M Contig791:1649 G G G C C G G G
        
        """
        self.col_name2index = {}	#col_name2index does not include the first two columns, starting only from samples
            #same as sampleID2index
        headerLineData = next(self)
        if headerLineData.lineIndicator=='I':
            self.header = headerLineData.genotypeList
            for i in range(2,len(self.header), self.ploidy):
                self.sampleIDList.append(self.header[i])
        else:
            if headerLineData.lineIndicator=='M':	#there is no header line. roll back to beginning
                self.csvFile.seek(0)
            #make up a header
            self.header = []
            for i in range(0, len(headerLineData.genotypeList),self.ploidy):
                sampleID = 'sampleColumn%s'%(i)
                self.sampleIDList.append(sampleID)
                for j in range(self.ploidy):	#same samples in multiple columns
                    self.header.append(sampleID)
        for i in range(len(self.header)):
            sampleID = self.header[i]
            if sampleID not in self.col_name2index:
                self.col_name2index[sampleID] = i	# only first haplotype is recorded if ploidy >1
            #self.col_name2index[sampleID].append(i)
            #the index corresponds to genotypeLikelihoodList in next().
            #so it starts from 0
        return self.col_name2index
    
    def getHaplotypeListOfOneSample(self, sampleID=None):
        """
        2013.05.23
            return a list of haplotypes for one individual, from self.snpData
            the number of haplotypes fetched depends on self.ploidy
        """
        self._completeSNPData()
        haplotypeList = []
        firstHaplotypeIndex = self.snpData.col_id2col_index.get(sampleID)
        for j in range(firstHaplotypeIndex, firstHaplotypeIndex+self.ploidy):
            #
            haplotypeList.append(self.snpData.data_matrix[:,j])
        return haplotypeList
    
    def getGenotypeOfOneSampleFromOneLocus(self, oneLocus=None, sampleID=None):
        """
        2013.05.06
            oneLocus is output of next(), upon iterating through the file
        """
        try:
            sampleStartIndex = self.getSampleIndexGivenSampleID(sampleID)
        except:
            sys.stderr.write('Except type: %s\n'%repr(sys.exc_info()))
            import traceback
            traceback.print_exc()
            sys.stderr.write("  sampleID %s not in file.\n"%(sampleID))
            raise
        genotypeList = oneLocus.genotypeList[sampleStartIndex:sampleStartIndex+self.ploidy]
        return genotypeList
    
    def getGenotypeOfOneSampleOneLocus(self, sampleID=None, locusID=None):
        """
        2013.06.05
            get genotype from self.snpData
        """
        self._completeSNPData()
        
        sampleStartIndex = self.getSampleIndexGivenSampleID(sampleID)
        locusIndex = self.snpData.row_id2row_index.get(locusID)
        sampleIndexList = []
        for i in range(sampleStartIndex, sampleStartIndex+self.ploidy):
            sampleIndexList.append(i)
        return self.snpData.data_matrix[locusIndex, sampleIndexList]
    
    def __next__(self):
        try:
            row = next(self.csvFile)
        except:
            raise StopIteration
        if not self.isRealCSV:
            row = row.strip().split()
        lineIndicator, markerID = row[0:2]
        
        return PassingData(markerID=markerID, lineIndicator=lineIndicator, genotypeList=row[2:])
    
    def readInAllHaplotypes(self, ):
        """
        2013.05.23
            this function reads in the entire file and fills in these data structures
                self.sampleIDList
                self.locusIDList
                self.snpData
        """
        sys.stderr.write("Reading in all haplotypes from %s ..."%(self.path))
        if self.inputFile:
            if self.inputFile.tell()!=0:
                self.inputFile.seek(0)
        self.constructColName2IndexFromHeader()
        dataMatrix = []
        for oneLocusData in self:
            self.locusIDList.append(oneLocusData.markerID)
            dataMatrix.append(oneLocusData.genotypeList)
        
        self.snpData = SNPData(row_id_ls=self.locusIDList, col_id_ls=self.sampleIDList, \
                            data_matrix=dataMatrix, ploidy=self.ploidy, matrix_orientation=2)
            #matrix_orientation=2, rows are loci, columns are samples
        self.sampleID2index = self.snpData.col_id2col_index
        sys.stderr.write("%s samples, %s loci.\n"%(len(self.sampleIDList), len(self.locusIDList)))
        
    def addOneIndividual(self, sampleID=None, locusIDList=None, haplotypeList=None):
        """
        2013.05.30
            this is for writing mode
        """
        self.sampleIDList.append(sampleID)
        self.haplotypeMatrix.append(haplotypeList)
        if locusIDList is not None:
            if not self.locusIDList:
                self.locusIDList = locusIDList
            elif self.locusIDList!=locusIDList:
                sys.stderr.write("Error: this individual %s has a different list of locus ID than this file (%s) already has %s.\n"%\
                                (sampleID, self.path, self.locusIDList))
                raise
            
    def _consolidateHaplotypeMatrixIntoSNPData(self, ):
        """
        2013.06.04
            if self.sampleIDList, self.locusIDList, self.haplotypeMatrix are all available (from self.addOneIndividual()),
                but self.snpData is not
            the key step is to turn a list of lists into a numpy 2D array
        """
        sys.stderr.write("Consolidating sampleIDList, locusIDList, haplotypeMatrix into SNPData ...")
        dataMatrix = numpy.array(self.haplotypeMatrix).transpose()	#transpose to make a column corresponding to haplotype 
        self.snpData = SNPData(row_id_ls=self.locusIDList, col_id_ls=self.sampleIDList, \
                            data_matrix=dataMatrix, ploidy=self.ploidy, matrix_orientation=2)
            #matrix_orientation=2, rows are loci, columns are samples
        self.sampleID2index = self.snpData.col_id2col_index
        sys.stderr.write("%s samples, %s loci.\n"%(len(self.sampleIDList), len(self.locusIDList)))
    
    def writeDataToDisk(self):
        """
        2013.05.30
        """
        sys.stderr.write("Writing data to file %s ..."%(self.path))
        if self.snpData is None:
            if self.sampleIDList and self.locusIDList and self.haplotypeMatrix:
                self._consolidateHaplotypeMatrixIntoSNPData()
            else:
                sys.stderr.write("Error: either sampleIDList (%s) or locusIDList (%s) or haplotypeMatrix is None/empty.\n"%\
                                (repr(self.sampleIDList[:3]), repr(self.locusIDList[:3]),  ))
                raise
        self.writerow(['I', 'id'] + self.sampleIDList)
        for i in range(len(self.locusIDList)):
            locusID = self.locusIDList[i]
            genotypeList = self.haplotypeMatrix[i:]
            oneRow = ['M', locusID] + genotypeList
            self.writerow(oneRow)
        sys.stderr.write(" .\n")
    
if __name__ == '__main__':
    main_class = BeagleGenotypeFile
    po = ProcessOptions(sys.argv, main_class.option_default_dict, error_doc=main_class.__doc__)
    instance = main_class(**po.long_option2value)
    instance.run()


"""
2008-10-01 module related to Genome
"""

import os,sys
import logging
from palos.utils import PassingData

class GeneModel(object):
	def __init__(self, **keywords):
		"""
		2008-10-01
			moved from transfac.src.GenomeDB in order for the return of
				GenomeDB.get_gene_id2model() to be pickled independent of transfac.src
		2008-10-01
			a class to hold all stuff related to a gene (Gene+EntrezgeneMapping)
			it's hierarchical. Its gene_commentary_ls contains also GeneModel.
		"""
		for argument_key, argument_value in keywords.items():
			setattr(self, argument_key, argument_value)
		if not hasattr(self, 'gene_commentary_ls'):
			self.gene_commentary_ls = []

class fasta_block_iterator(object):
	'''
	2011-1-5 moved from Transfac.src.transfacdb
	09-13-05
		fasta format iterator
		a little bit tricky, '>', the block starter is used as a tokenizer
	2006-09-01
		it seems 'for line in self.inf' doesn't work on hpc-cmb.
		check https://dl403k-1.cmb.usc.edu/log/hpc-cmb
	'''
	def __init__(self, inf):
		self.inf = inf
		self.block = ''
		self.previous_line = ''
	def __iter__(self):
		return self
	def next(self):
		self.read()
		return self.block
	def read(self):
		self.block = self.previous_line	#don't forget the starting line
		line = self.inf.readline()
		while(line):
			if line[0]=='>':
				self.previous_line = line
				if self.block:	#not the first time
					break
				else:	#first time to read the file, block is still empty
					self.block += line
			else:
				self.block += line
			line = self.inf.readline()
		if self.block==self.previous_line:	#nothing new into the block
			raise StopIteration

class LargeFastaFileTraverse:
	"""
	2010-1-5
	"""
	def __init__(self):
		pass
	
	def traverse(self, input_dir, headerFunctor=None, seqlineFunctor=None):
		"""
		2011-1-5
			orignal idea is in transfac/src/chromosome_fasta2db.py.
			Each fasta file in input_dir could contain many fasta blocks.
		"""
		import re
		p_chromosome = re.compile(r'chromosome (\w+)[,\n\r]?')
		#the last ? means [,\n\r] is optional
		files = os.listdir(input_dir)
		no_of_total_files = len(files)
		
		for i in range(no_of_total_files):
			fname = files[i]
			input_fname = os.path.join(input_dir, fname)
			sys.stderr.write("\t %s/%s %s ..."%(i+1, no_of_total_files, fname))
			if fname[-2:]=='gz':
				import gzip
				inf = gzip.open(input_fname)
			else:
				inf = open(input_fname)
			line = inf.readline()
			new_fasta_block = 1
			#'line' is not enough to stop the 'while' loop.
			# after the file reading is exhausted by "for line in inf:",
			#  'line' still contains the stuff from the last line.
			no_of_fasta_blocks = 0
			while line and new_fasta_block:
				new_fasta_block = 0
				#set it to 0, assuming only one fasta block, change upon new fasta block
				if line[0]!='>':	#not fasta block header
					for line in inf:
						#exhaust this fasta block as it's not what's wanted.
						if line[0]=='>':
							new_fasta_block = 1
							break	#start from while again
					continue
				
				headerFunctor(line)
				"""
				# 2010-1-5 an example header function which parses the title of the fasta block 
				#possible header lines:
				#>gi|51511461|ref|NC_000001.8|NC_000001 Homo sapiens chromosome 1, complete sequence
				#>gi|186497660|ref|NC_003070.6| Arabidopsis thaliana chromosome 1, complete sequence
				#>gi|26556996|ref|NC_001284.2| Arabidopsis thaliana mitochondrion, complete genome
				#>gi|115442598|ref|NC_008394.1| Oryza sativa (japonica cultivar-group) genomic DNA, chromosome 1
				header = line[1:-1]	#discard '>' and '\n'
				header = header.split('|')
				
				if tax_id is None:
					_tax_id = FigureOutTaxID_ins.returnTaxIDGivenSentence(header[4])
					if not _tax_id:
						_tax_id = 'null'
				else:
					_tax_id = tax_id
				
				if p_chromosome.search(header[4]) is not None:
					chromosome = p_chromosome.search(header[4]).groups()[0]
				elif header[4].find('mitochondrion')!=-1:
					chromosome = 'mitochondrion'
				elif header[4].find('chloroplast')!=-1:
					chromosome = 'chloroplast'
				else:	#something else, take the whole before ','
					chromosome = header[4].split(',')[0]
				
				outf.write(">chr%s\n"%(chromosome))
				"""
				for line in inf:
					if line[0]=='>':
						
						new_fasta_block = 1
						break	#start from while again
					else:
						seqlineFunctor(line)
			sys.stderr.write("\n")

import re
chr_pattern = re.compile(r'([a-zA-Z]+[\dXY]+)[._\-:]*')
#the last - has special meaning in [] when it's not the last character. 
#contig_id_pattern = re.compile(r'Contig(\d+)[._\-:]*')
#contig_id_pattern = re.compile(r'Scaffold(\d+)[._\-:]*')
# #2013.3.19 new vervet ref is scaffold-based.
contig_id_pattern = re.compile(r'[CcS][a-zA-Z]+([\dXY]+)[._\-:]*')
#2013.07.16 new chromosome (Chlorocebus aethiops) vervet ref, add X, Y
chr_start_stop_pattern = re.compile(r'([a-zA-Z]+[\dXY]+)_(\d+)_(\d+)[._\-:]*')
#the last - has special meaning in [] when it's not the last character. 

def getContigIDFromFname(filename):
	"""
	2012.7.14 copied from  pymodule.pegasus.AbstractNGSWorkflow
	2011-10-20
		
		If filename is like .../Contig0.filter_by_vcftools.recode.vcf.gz,
			It returns "0", excluding the "Contig".
			If you want "Contig" included, use getChrIDFromFname().
		If search fails, it returns the prefix in the basename of filename.
	"""
	contig_id_pattern_sr = contig_id_pattern.search(filename)
	if contig_id_pattern_sr:
		contig_id = contig_id_pattern_sr.group(1)
	else:
		contig_id = os.path.splitext(os.path.split(filename)[1])[0]
	return contig_id

def getChrFromFname(filename):
	"""
	2012.7.14 copied from  pymodule.pegasus.AbstractNGSWorkflow
	2011-10-20
		filename example: Contig0.filter_by_vcftools.recode.vcf.gz
			It returns "Contig0".
			If you want just "0", use getContigIDFromFname().
		If search fails, it returns the prefix in the basename of filename.
	"""
	searchResult = chr_pattern.search(filename)
	if searchResult:
		chromosome = searchResult.group(1)
	else:
		chromosome = os.path.splitext(os.path.split(filename)[1])[0]
	return chromosome

def parseChrStartStopFromFilename(filename=None, chr2size=None,
	defaultChromosomeSize=10000000000):
	"""
	2013.09.18
		#10000000000 is used when filename contains data from a whole chromosome
		#  and chr2size is not available or not containing chromosome
		make it very big so that it could be intersected with any interval
			from any chromosome.
	"""
	searchResult = chr_start_stop_pattern.search(filename)
	if searchResult:
		chromosome = searchResult.group(1)
		start = int(searchResult.group(2))
		stop = int(searchResult.group(3))
	else:	#try
		chromosome = getChrFromFname(filename=filename)
		start =1
		if chr2size is not None:
			stop = chr2size.get(chromosome, defaultChromosomeSize)
		else:
			stop = defaultChromosomeSize
	return PassingData(chromosome=chromosome, start=start, stop=stop)

class IntervalData(PassingData):
	"""
	Access .overlapInterval for most activities, as it's either equal or
		bigger than .interval.
	
	required keywords: chr or chromosome, chromosomeSize, start, stop
		overlapStart and overlapStop are optional.
	
	2013.09.05 a more dynamic way to store intervals
		
		
	"""
	def __init__(self, chr=None, chromosome=None, chromosomeSize=None,
		start=None, stop=None, overlapStart=None, overlapStop=None, **keywords):
		PassingData.__init__(self, chr=chr, chromosome=chromosome,
			chromosomeSize=chromosomeSize, \
			start=start, stop=stop, 
			overlapStart=overlapStart, overlapStop=overlapStop, **keywords)
		if not hasattr(self, 'file'):
			self.file = None
		if not hasattr(self, 'jobLs'):
			self.jobLs = []
		if self.chr is None and self.chromosome:
			self.chr = self.chromosome
		elif self.chr and self.chromosome is None:
			self.chromosome = self.chr
		
		if self.overlapStart is None:
			self.overlapStart = self.start
		
		if self.overlapStop is None:
			self.overlapStop = self.stop
		
		self.subIntervalLs = []
		self.subIntervalLs.append((self.overlapStart, self.overlapStop))
	
	def unionOneIntervalData(self, otherIntervalData=None):
		"""
		"""
		if self.chromosome!=otherIntervalData.chromosome:
			logging.error("This interval %s is trying to union an interval from a different chromosome %s."%\
				(self.interval, otherIntervalData.interval))
			raise
		self.subIntervalLs.append((otherIntervalData.overlapStart, otherIntervalData.overlapStop))
		self.subIntervalLs.sort()
		self.start = min(self.start, otherIntervalData.start)
		self.stop = max(self.stop, otherIntervalData.stop)
		self.overlapStart = min(self.overlapStart, otherIntervalData.overlapStart)
		self.overlapStop = max(self.overlapStop, otherIntervalData.overlapStop)
		
		
	@property
	def span(self):
		deltaLs = [a[1]-a[0]+1 for a in self.subIntervalLs]
		return sum(deltaLs)
	
	@property
	def overlapIntervalFileBasenameSignature(self):
		return '%s_%s_%s'%(self.chromosome, self.overlapStart, self.overlapStop)
	
	@property
	def intervalFileBasenameSignature(self):
		return '%s_%s_%s'%(self.chromosome, self.start, self.stop)
		
		
	@property
	def overlapInterval(self):
		return "%s:%s-%s"%(self.chromosome, self.overlapStart, self.overlapStop)
	@property
	def interval(self):
		return "%s:%s-%s"%(self.chromosome, self.start, self.stop)
	@property
	def parentInterval(self):
		return self.interval




#!/usr/bin/env python3
"""
2012.12.15 table-data stored in pytables.
i.e.
    reader = PyTablesMatrixFile(path=filename, mode='r')
    reader = PyTablesMatrixFile(filename, mode='r')
    for row in reader:
        ...
    tableObject = reader.getTableObject(tableName=tableName)
    for row in tableObject:
        ...
    
    dtypeList = [('locus_id','i8'),('chr', HDF5MatrixFile.varLenStrType), ('start','i8'), ('stop', 'i8'), \
                ('score', 'f8'), ('MAC', 'i8'), ('MAF', 'f8')]
    headerList = [row[0] for row in dtypeList]
    dtype = numpy.dtype(dtypeList)
    
    writer = PyTablesMatrixFile(path=filename, mode='w', dtype=dtype)
    writer = PyTablesMatrixFile(filename, mode='w', dtype=dtype)
    
    if writer:
        tableObject = writer.createNewTable(tableName=tableName, dtype=dtype)
        tableObject.setColIDList(headerList)
    elif outputFname:
        writer = PyTablesMatrixFile(outputFname, mode='w', dtype=dtype, tableName=tableName)
        writer.writeHeader(headerList)
        tableObject = writer.getTableObject(tableName=tableName)
    cellList = []
    for data_obj in self.data_obj_ls:
        dataTuple = self._extractOutputRowFromDataObject(data_obj=data_obj)
        cellList.append(dataTuple)
    tableObject.writeCellList(cellList)
    if closeFile:
        writer.close()
        del writer
    
    #each number below is counting bytes, not bits
    dtypeList = [('locus_id','i8'),('chromosome', HDF5MatrixFile.varLenStrType), ('start','i8'), ('stop', 'i8'), \
                ('score', 'f8'), ('MAC', 'i8'), ('MAF', 'f8')]
    if writer is None and filename:
        writer = PyTablesMatrixFile(filename, mode='w', dtypeList=dtypeList, tableName=tableName)
        tableObject = writer.getTableObject(tableName=tableName)
    elif writer:
        tableObject = writer.createNewTable(tableName=tableName, dtypeList=dtypeList)

"""

import sys, os
import csv
import tables
import numpy
from palos.utils import PassingData, PassingDataList
from palos.ProcessOptions import  ProcessOptions
from palos.io.MatrixFile import MatrixFile
from palos.io.HDF5MatrixFile import HDF5MatrixFile, YHTableInHDF5Group, addAttributeDictToYHTableInHDF5Group 
from tables.file import _checkfilters

class YHTable(tables.Table, YHTableInHDF5Group):
    """
    2012.12.16 API is very similar to HDF5MatrixFile
    2012.12.16 adapted from http://pytables.github.com/cookbook/simple_table.html
    """
    #mimics the sqlalchemy
    query = tables.Table.read_where
    
    def __init__(self, parentNode=None, tableName=None,\
                description=None, rowDefinition=None,\
                title='', filters=None, \
                expectedrows=52000, chunkshape=None, byteorder=None, **keywords):
        """
    description
          An IsDescription subclass or a dictionary where the keys are the field
          names, and the values the type definitions. In addition, a pure NumPy
          dtype is accepted.  And it can be also a recarray NumPy object,
          RecArray numarray object or NestedRecArray. If None, the table metadata
          is read from disk, else, it's taken from previous parameters.
    rowDefinition is a backup of description, to make the class compatible with HDF5MatrixFile
        """
        self.tableName = tableName
        self.parentNode = parentNode
        
        if tableName in self.parentNode: # existing table
            description = None
        elif description is None: # pull the description from the attrs
            if rowDefinition is not None:	#a backup of description
                description = rowDefinition
            else:
                description = dict(self._get_description())
        
        if filters is None:
            filters = tables.Filters(complib="blosc", complevel=5, shuffle=True)
        
        #YHTableInHDF5Group.__init__(self, newGroup=0)
        #if self.mode=='w':
        tables.Table.__init__(self, self.parentNode, tableName,
                        description=description, title=title,
                        filters=filters,
                        expectedrows=expectedrows,
                        _log=False, **keywords)
        
        self._c_classId = self.__class__.__name__
        
        self._processRowIDColID()

    def _processRowIDColID(self):
        """
        2012.12.16 similar to SNPData.processRowIDColID()
        """
        self.rowIDList = []
        self.rowID2rowIndex = {}
        self.colIDList = []
        self.colID2colIndex = {}
        for i in range(len(self.colnames)):
            colID = self.colnames[i]
            self.colIDList.append(colID)
            self.colID2colIndex[colID] = i
        
        self.no_of_rows = self.nrows	#a counter used in self.writeOneCell, self.nrows does not get updated until flush.
        
        
    
    def _get_description(self):
        # pull the description from the attrs
        for attr_name in dir(self):
            if attr_name[0] == '_':
                continue
            try:
                attr = getattr(self, attr_name)
            except:
                continue
            if isinstance(attr, tables.Atom):
                yield attr_name, attr

    def insert_many(self, data_generator, attr=False):
        row = self.row
        cols = self.colnames
        if not attr:
            for d in data_generator:
                for c in cols:
                    row[c] = d[c]
                row.append()
        else:
            for d in data_generator:
                for c in cols:
                    row[c] = getattr(d, c)
                row.append()
        self.flush()
    
    def setColIDList(self, colIDList=None, **keywords):
        """
        """
        pass
    
    def setRowIDList(self, rowIDList=None, **keywords):
        pass
    
    def writeHeader(self, headerList=None, tableIndex=None, tableName=None):
        """
        2012.11.16
            only the first group
        """
        pass
    
    def writeOneCell(self, oneCell=None, cellType=1):
        """
        #2013.1.9 make sure enough columns in oneCell
        2012.12.16
            each cell is either a tuple/list or object with column-name attributes
            cellType
                1: list or tuple, in the order of table columns, not including first "id" column
                2: object with attributes whose names are same as those of the table columns
                3: a dictionary with key=column-name, value=column-value, access by oneCell.get(colname, None)
                4: raw dictionary. access by oneCell[colname]
        """
        row = self.row
        for i in range(len(self.colnames)):	#assuming data in oneCell is in the same order as tableObject.colnames
            colname = self.colnames[i]
            if colname=='id':	#auto-increment the ID column if it exists
                self.no_of_rows += 1
                row[colname] = self.no_of_rows
            else:
                if cellType==1:	#list/tuple
                    if "id" in self.colinstances:	#colinstances maps the name of a column to its Column (see The Column class) 
                            #or Cols (see The Cols class) instance.
                        cellColIndex = i-1					#oneCell does not include id
                    else:	#no "id" column, so same index
                        cellColIndex = i
                    if cellColIndex<len(oneCell):	#2013.1.9 make sure enough columns in oneCell
                        value = oneCell[cellColIndex]
                    else:
                        continue
                elif cellType==2:	#object with attributes
                    value = getattr(oneCell, colname, None)
                elif cellType==3:	#dictionary
                    value = oneCell.get(colname, None)
                elif cellType==4:	#canonical dictionary
                    value = oneCell[colname]
                else:
                    value = None
                if value is not None:
                    row[colname] = value
        row.append()
    
    def appendOneRow(self, oneCell=None, cellType=1):
        """
        2013.1.9
        """
        self.writeOneCell(oneCell=oneCell, cellType=cellType)
    
    def writeCellList(self, cellList=None, cellType=1):
        """
        """
        for oneCell in cellList:
            self.writeOneCell(oneCell, cellType=cellType)
        self.flush()
    
    def getCellDataGivenRowColID(self, rowID=None, colID=None):
        """
        """
        pass
        """
        rowIndex = self.getRowIndex(rowID)
        colIndex = self.getColIndex(colID)
        
        cellData = None
        if rowIndex is not None and colIndex is not None:
            cellData = self.[rowIndex][colIndex]
        return cellData
        """
    
    def addAttribute(self, name=None, value=None, overwrite=True):
        """
        
        """
        if hasattr(self._v_attrs, name):
            sys.stderr.write("Warning: h5Group %s already has attribute %s=%s.\n"%(self.name, name, value))
            if overwrite:
                setattr(self._v_attrs, name, value)
            else:
                return False
        else:
            setattr(self._v_attrs, name, value)
        #pass the HDF5Group attributes to this object itself , it ran into "can't set attribute error". conflict with existing property
        #object.__setattr__(self, name, value)
        #setattr(self, name, value)
        return True
    
    def getAttribute(self, name=None, defaultValue=None):
        return getattr(self._v_attrs, name, defaultValue)
    
    def getAttributes(self):
        dc = {}
        for attributeName in self._v_attrs._f_list():
            dc[attributeName] = getattr(self._v_attrs, attributeName)
        return dc
    
    #2102.12.20 overwrite YHTableInHDF5Group.next()
    #next = tables.Table.next	tables.Table doesn't have next
    
    
    def close(self):
        """
        2012.12.20 do nothing
        """
        pass


class YHSingleTableFile(YHTable):
    """
    2012.12.16 API is very similar to HDF5MatrixFile
    2012.12.16 adapted from http://pytables.github.com/cookbook/simple_table.html
    """
    #mimics the sqlalchemy	
    def __init__(self, path=None, mode='r', \
                groupName=None, tableName=None,\
                description=None,
                title='', filters=None, rowDefinition=None,\
                expectedrows=512000, **keywords):
        """
        rowDefinition is backup of description, to make it compatible with HDF5MatrixFile
        """
        self.path = path
        self.mode = mode
        self.groupName = groupName
        self.tableName = tableName
        
        self.hdf5File = tables.openFile(path, mode)
        self.uservars = None
        
        if groupName is None:
            groupName = 'default'
        
        self.parentNode = self.hdf5File._getOrCreatePath('/' + groupName, True)
        
        YHTable.__init__(self, self.parentNode, tableName,
                        description=description, rowDefinition=rowDefinition, \
                        title=title,
                        filters=filters,
                        expectedrows=expectedrows,
                        _log=False, **keywords)
        
        self._c_classId = self.__class__.__name__
    
    def getTableObject(self, tableIndex=None, tableName=None, groupName=None):
        """
        2012.12.18 compatible with HDF5MatrixFile
        """
        return self

class transplant:
    """
    2012.12.20 from http://code.activestate.com/recipes/81732/
    """
    def __init__(self, method, host, method_name=None):
        self.host = host
        self.method = method
        setattr(host, method_name or method.__name__, self)	#calling method_name is effectively calling this transplant class

    def __call__(self, *args, **kwargs):
        nargs = [self.host]
        nargs.extend(args)
        return apply(self.method, nargs, kwargs)

def funcToMethod(func, newClass, method_name=None, instance=None):
    """
    2012.12.20 from http://code.activestate.com/recipes/81732/,
        to attach func to newClass with new method_name and instance
    """
    import new
    method = new.instancemethod(func, instance, newClass)
    if not method_name:
        method_name=func.__name__
    instance.__dict__[method_name]=method

#def turnTableIntoYHTable():
#	for functionName in ['getAttributes', 'getAttribute', 'addAttribute', 'addAttributeDict']:
#		function = getattr(YHTable, functionName)
#		funcToMethod(function, tables.Table, functionName)

#turnTableIntoYHTable()


def armTableObjectWithYHTableMethods(tableObject=None):
    """
    2012.12.20 function to arm a tableObject (read from a pytables file) with functions from YHTable and its parental class YHTableInHDF5Group
        almost like type-cast (python doesn't support type cast) 
    """
    import types
    for functionName in ['_processRowIDColID', 'addAttribute', 'addAttributeDict', 'addObjectAttributeToSet', \
                        'addObjectListAttributeToSet', 'close', 'constructColName2IndexFromHeader', \
                        'getAttribute', 'getAttributes',  'getCellDataGivenRowColID', \
                        'getColIndex', 'getColIndexGivenColHeader', \
                        'getListAttributeInStr', \
                        'getRowIndex', \
                        'writeOneCell', 'writeCellList', 'setColIDList', 'setRowIDList']:
        if functionName in tableObject.__dict__:	#make sure not overwriting tables.Table's own function
            continue 
        if functionName in YHTable.__dict__:	#YHTable has precedence
            ClassToBorrow = YHTable
        elif functionName in YHTableInHDF5Group.__dict__:
            ClassToBorrow = YHTableInHDF5Group
        else:
            continue
        function = ClassToBorrow.__dict__[functionName]
        
        #function = getattr(YHTable, functionName)
        ## use getattr() from above to get the function will result in this error:
        ###"TypeError: unbound method getAttribute() must be called with YHTable instance as first argument (got Table instance instead)"
        
        tableObject.__dict__[functionName] = types.MethodType(function, tableObject, tableObject.__class__)
        
        #transplant(function, tableObject)	#another way of doing this, won't work because these functions are still bound to their old class
        
        #function.im_class = tables.Table	#read-only attributes for bound methods
        #func.im_func=fun
        #function.im_self = tableObject
        #funcToMethod(function, tables.Table, functionName, instance=tableObject)
    tableObject._processRowIDColID()
    return tableObject

class YHFile(tables.File, HDF5MatrixFile):
    """
    2012.12.18 not ready. a pytables file that could contain multiple tables.
    2012.12.16 API is very similar to HDF5MatrixFile
    i.e.
    
    #open a file with a defined table structure
    writer = YHFile(self.outputFname, mode='w', rowDefinition=CountAssociationLocusTable)
    ...
    
    #read a file
    reader = YHFile(path, mode='r')
    ...
    
    #open a file without passing the table structure.
    writer = YHFile(self.outputFname, mode='w')
    # create a table with proper structure
    writer.createNewTable(rowDefinition=CountAssociationLocusTable)
    ...
    """
    def __init__(self, path=None, mode='r', \
                tableName=None, groupNamePrefix='group', tableNamePrefix='table',\
                rowDefinition=None, filters=None, expectedrows=500000, \
                autoRead=True, autoWrite=True, \
                debug=0, report=0, **keywords):
        self.path = path
        self.header = None
        self.mode = mode
        self.tableName = tableName
        self.groupNamePrefix = groupNamePrefix
        self.tableNamePrefix = tableNamePrefix
        self.rowDefinition = rowDefinition
        self.expectedrows = expectedrows
        self.autoRead = autoRead	#whether to automatically read in data if mode is in reading mode
        self.autoWrite = autoWrite	#whether to automatically create table objects if mode is in writing mode
        
        self.debug = debug
        self.report = report
        
        self.combinedColIDList = None	#same as header
        self.combinedColID2ColIndex = None
        
        #to get table objects fast, without retrieving them from their parental groups.
        self.tableObjectList = []
        self.tablePath2Index = {}
        
        #self.hdf5File = tables.openFile(self.path, self.mode)
        #self.root = self.hdf5File.root
        if filters is None:
            filters = tables.Filters(complib="blosc", complevel=5, shuffle=True)
        
        tables.File.__init__(self, self.path, mode=self.mode, title='', rootUEP='/', filters=filters,\
                            **keywords)
        
        if self.mode=='r' or self.mode=='a':
            self._setupCombinedColumnIDMapper()
            if self.autoRead:
                self._readInData(tableName=self.tableName)
        
        if self.autoWrite and self.mode=='w' and self.rowDefinition is not None:
            self.createNewTable(tableName=self.tableName, rowDefinition=self.rowDefinition, expectedrows=self.expectedrows)
        
        self._c_classId = self.__class__.__name__
    
    def _createNewGroup(self, groupName=None, **keywords):
        """
        2012.12.16 this would also create a new table object within the newly-created group
            return tableObject
            ** deprecated
        """
        if groupName is None or self.root.__contains__(groupName):
            groupName = self._getNewGroupName()
        groupObject = self.createGroup(self.root, groupName)
        return groupObject
    
    def _appendNewTable(self, tableObject=None):
        """
        2012.12.20
        """
        if tableObject is not None:
            tablePath = tableObject._v_pathname	##the tableName preceds with "/"
            if tablePath in self.tablePath2Index:	 
                sys.stderr.write("ERROR, table %s already in self.tablePath2Index, index=%s.\n"%(tablePath,\
                                                                            self.tablePath2Index.get(tablePath)))
                sys.exit(3)
            self.tablePath2Index[tablePath] = len(self.tablePath2Index)
            self.tableObjectList.append(tableObject)
    
    def _fetchTableObject(self, tablePath):
        """
        2012.12.20
        """
        tableObject = None
        if tablePath in self.tablePath2Index:
            tableObject = self.tableObjectList[self.tablePath2Index.get(tablePath)]
        return tableObject
    
    def createNewTable(self, tableName=None, rowDefinition=None, title="",
                    filters=None, expectedrows=50000,
                    chunkshape=None, byteorder=None):
        """
        2012.12.20 this would also create a new group object if no group exists.
            symlink to createTable()
        """
        groupObject = self._getGroupObject()
        if not groupObject:	#create a new one if it's not available.
            groupName = self._getNewGroupName()
            groupObject = self.createGroup(self.root, groupName)
        
        if tableName is None:
            tableName = self._getNewTableName(groupObject=groupObject)
        if rowDefinition is None:
            rowDefinition = self.rowDefinition
        return self.createTable(where=groupObject, tableName=tableName, description=rowDefinition, \
                            rowDefinition=rowDefinition, title=title, filters=filters, expectedrows=expectedrows, \
                            chunkshape=chunkshape, byteorder=byteorder, createparents=False)
    
    def createTable(self, where=None, tableName=None, description=None, rowDefinition=None, title="",
                    filters=None, expectedrows=50000,
                    chunkshape=None, byteorder=None,
                    createparents=False):
        """
        2012.12.20 overwrite the parent (stock) function
        
        Create a new table with the given name in where location.

        Parameters
        ----------
        where : str or Group
            The parent group from which the new table will hang. It can be a
            path string (for example '/level1/leaf5'), or a Group instance
            (see :ref:`GroupClassDescr`).
        name : str
            The name of the new table.
        description : Description
            This is an object that describes the table, i.e. how
            many columns it has, their names, types, shapes, etc.  It
            can be any of the following:

                * *A user-defined class*: This should inherit from the
                  IsDescription class (see :ref:`IsDescriptionClassDescr`) where
                  table fields are specified.
                * *A dictionary*: For example, when you do not know beforehand
                  which structure your table will have).
                * *A Description instance*: You can use the description
                  attribute of another table to create a new one with the same
                  structure.
                * *A NumPy dtype*: A completely general structured NumPy dtype.
                * *A NumPy (structured) array instance*: The dtype of this
                  structured array will be used as the description.  Also, in
                  case the array has actual data, it will be injected into the
                  newly created table.
                * *A RecArray instance (deprecated)*: Object from the numarray
                  package.  This does not give you the possibility to create a
                  nested table.  Array data is injected into the new table.
                * *A NestedRecArray instance (deprecated)*: If you want to have
                  nested columns in your table and you are using numarray, you
                  can use this object. Array data is injected into the new
                  table.

        title : str
            A description for this node (it sets the TITLE HDF5 attribute
            on disk).
        filters : Filters
            An instance of the Filters class (see :ref:`FiltersClassDescr`) that
            provides information about the desired I/O filters to be applied
            during the life of this object.
        expectedrows : int
            A user estimate of the number of records that will be in the table.
            If not provided, the default value is EXPECTED_ROWS_TABLE (see
            :file:`tables/parameters.py`). If you plan to create a bigger table
            try providing a guess; this will optimize the HDF5 B-Tree creation
            and management process time and memory used.
        chunkshape
            The shape of the data chunk to be read or written in a single HDF5
            I/O operation. Filters are applied to those chunks of data. The rank
            of the chunkshape for tables must be 1. If None, a sensible value
            is calculated based on the expectedrows parameter (which is
            recommended).
        byteorder : str
            The byteorder of data *on disk*, specified as 'little' or 'big'.
            If this is not specified, the byteorder is that of the platform,
            unless you passed an array as the description, in which case
            its byteorder will be used.
        createparents : bool
            Whether to create the needed groups for the parent path to exist
            (not done by default).

        See Also
        --------
        Table : for more information on tables

        """
        if rowDefinition is None:
            rowDefinition = self.rowDefinition
        if description is None:	#2012.12.20 rowDefinition is backup of description
            description = rowDefinition
        parentNode = self._getOrCreatePath(where, createparents)
        _checkfilters(filters)
        tableObject =  YHTable(parentNode=parentNode, tableName=tableName,
                     description=description, title=title,
                     filters=filters, expectedrows=expectedrows,
                     chunkshape=chunkshape, byteorder=byteorder)
        if tableObject._v_pathname not in self.tablePath2Index:
            self._appendNewTable(tableObject)
        else:
            sys.stderr.write("Error: table path name %s already in this PyTables file.\n"%(tableObject._v_pathname))
            sys.exit(3)
        return tableObject
    
    def _getNewGroupName(self, groupNamePrefix=None):
        """
        2012.11.19
        """
        if not groupNamePrefix:
            groupNamePrefix = self.groupNamePrefix
        i = len(self.root._v_groups)
        groupName = "%s%s"%(groupNamePrefix, i)
        while self.root.__contains__(groupName):	#stop until a unique name shows up
            i += 1
            groupName = "%s%s"%(groupNamePrefix, i)
        return groupName

    def _getNewTableName(self, tableNamePrefix=None, groupObject=None):
        """
        2012.12.15
        """
        if groupObject is None:
            groupObject = self._getGroupObject()
        if self.tableName:
            tableName = self.tableName
        else:
            if not tableNamePrefix:
                tableNamePrefix = self.tableNamePrefix
            i = len(groupObject._v_leaves)
            tableName = "%s%s"%(tableNamePrefix, i)
        while groupObject.__contains__(tableName):	#stop until a unique name shows up
            i += 1
            tableName = "%s%s"%(tableNamePrefix, i)
        return tableName
    
    def _readInData(self, tableName=None, tableObject=None):
        """
        2012.12.16
        """
        pass
    
    def _setupCombinedColumnIDMapper(self,):
        """
        2012.12.16
            use the first table only
        """
        self.combinedColID2ColIndex = {}
        tableObject = self.getTableObject()
        self.header = []
        self.combinedColIDList = self.header
        for colID in tableObject.colnames:
            if colID in self.combinedColID2ColIndex:
                sys.stderr.write("Error: column ID %s already used in column %s.\n"%(colID, self.combinedColID2ColIndex.get(colID)))
                sys.exit(3)
            else:
                self.combinedColID2ColIndex[colID] = len(self.combinedColID2ColIndex)
                self.header.append(colID)
        return self.combinedColID2ColIndex
    """
    def constructColName2IndexFromHeader(self):
        #2012.11.22 overwrite parent function
        return self.combinedColID2ColIndex
    
    def getColIndexGivenColHeader(self, colHeader=None):
        #2012.11.15
        #	this is from the combined column header list.
        return self.combinedColID2ColIndex.get(colHeader)
    """
    
    def _getGroupObject(self, groupIndex=None, groupName=None):
        """
        """
        groupObject = None
        if groupName:
            groupObject = self.getNode("/", name=groupName, classname='Group')
        else:	#return first group
            if len(self.root._v_groups.values())>0:
                groupObject = self.root._v_groups.values()[0]	#first group
        return groupObject
    
    def getTableObject(self, tableIndex=None, tableName=None, groupName=None):
        """
        2012.12.16
        """
        if tableName is None and self.tableName:
            tableName = self.tableName
        if tableIndex is None and tableName is None and self.tableObjectList:	#return the 1st one.
            return self.tableObjectList[0]
        groupObject = self._getGroupObject(groupIndex=None, groupName=groupName)
        tableObject = None	
        if tableName:
            tablePath = groupObject._v_pathname + '/' + tableName
            if tablePath in self.tablePath2Index:	#already cached
                return self._fetchTableObject(tablePath)
            else:
                tableObject = groupObject._f_getChild(tableName)
        else:
            tableObject = groupObject._v_leaves.values()[0]	#first leaf
        if not hasattr(tableObject, 'getAttribute'):
            tableObject = armTableObjectWithYHTableMethods(tableObject)
            self._appendNewTable(tableObject)
        return tableObject
    
    @property
    def nrows(self):
        """
        2012.12.21
        """
        tableObject = self.getTableObject()
        return tableObject.nrows
    
    def __iter__(self):
        """
        2012.12.23
        don't just "return self" and remove the next() because using "yield" is making a generator itself
            i.e.
            for row in yhFile:
                print row
            
            Python first calls obj=yhFile.__iter__(), then every "for" step, it calls next(obj).
             
            In this __iter__(), "yield something" returns a newObj = iter(something)
                (AKA a generator), the in each "for" step, python calls next(newObj).
        2012.12.23 have to overwrite tables.File's __iter__() otherwise it's returning hdf5 group objects (/, /group0, etc.) iteratively
        return self
    
    def next(self):
        2012.12.20 iterate through just the first table
        
        """
        
        row = None
        pdata = PassingDataList()
        tableObject = self.getTableObject(tableName=self.tableName)
        for row in tableObject:
            try:
                yield row	#yield row returns a generator
            except:
                raise StopIteration
        """
        #2012.12.16
        #	go through each leaf (table & array), iteratively.
        #	usually the file has only one leaf (=table)
        for groupObject in self.walkGroups("/"):
            for leafNode in  groupObject._f_walkNodes('Leaf'):
                for row in leafNode:
                    yield row
        """
        """
            for colID  in self.header:	#iteration over header is in the same order as the ascending order of colIndex
                colIndex = self.combinedColID2ColIndex.get(colID)
                if colIndex < len(row):
                    setattr(pdata, colID, row[colIndex])
        self.rowIndexCursor += 1
        return pdata
        """
    
    def writeHeader(self, headerList=None, tableIndex=None, tableName=None):
        """
        2012.11.16
            only the first group
        """
        pass
    
    def writeOneCell(self, oneCell=None, tableIndex=None, tableName=None, cellType=1):
        """
        2012.12.16
            mimic csv's writerow()
            cellType
                1: list or tuple, in the order of table columns, not including first "id" column
                2: object with attributes whose names are same as those of the table columns
                3: a dictionary with key=column-name, value=column-value
        """
        tableObject = self.getTableObject(tableIndex=tableIndex, tableName=tableName)
        tableObject.writeOneCell(oneCell, cellType=cellType)
        """
        row = tableObject.row
        for i in range(len(tableObject.colnames)):	#assuming data in oneCell is in the same order as tableObject.colnames
            colname = tableObject.colnames[i]
            row[colname] = oneCell[i]
        row.append()
        """
    
    def appendOneRow(self, oneCell=None, tableIndex=None, tableName=None, cellType=1):
        """
        2013.1.9
        """
        self.writeOneCell(oneCell=oneCell, tableIndex=tableIndex, tableName=tableName, cellType=cellType)
    
    def writeCellList(self, cellList=None, tableIndex=None, tableName=None, cellType=1):
        """
        """
        tableObject = self.getTableObject(tableIndex=tableIndex, tableName=tableName)
        tableObject.writeCellList(cellList, cellType=cellType)
        """
        for oneCell in cellList:
            self.writeOneCell(oneCell, tableIndex=tableIndex, tableName=tableName)
        tableObject.flush()
        """
    
    def setColIDList(self, colIDList=None, tableIndex=None, tableName=None):
        """
        """
        pass
        
    def setRowIDList(self, rowIDList=None, tableIndex=None, tableName=None):
        """
        """
        pass

def castPyTablesRowIntoPassingData(rowPointer=None):
    """
    2012.12.21 rowPointer from PyTables iteration is like a C pointer to the current row (no real content).
        need to convert it to a real object if you try to store its content in memory and use it later. 
    
    rowPointer has these methods: 
        ['__class__', '__contains__', '__delattr__', '__delitem__', '__doc__', '__format__', '__getattribute__', 
        '__getitem__', '__hash__', '__init__', '__iter__', '__new__', '__next__', '__pyx_vtable__', '__reduce__', 
        '__reduce_ex__', '__repr__', '__setattr__', '__setitem__', '__sizeof__', '__str__', '__subclasshook__', 
        '_fillCol', '_flushBufferedRows', '_flushModRows', '_getUnsavedNrows', '_iter', 'append', 'fetch_all_fields', 
        'next', 'nrow', 'table', 'update']
        
        However, errors like this crop up during copying of these rowPointer. 

  File "/usr/lib/python2.7/copy_reg.py", line 93, in __newobj__
    return cls.__new__(cls, *args)
  File "tableExtension.pyx", line 706, in tables.tableExtension.Row.__cinit__ (tables/tableExtension.c:6910)
TypeError: __cinit__() takes exactly 1 positional argument (0 given)

    2012.12.21 could not use PassingDataList. because of these errors.

  File "/usr/lib/python2.7/copy.py", line 257, in _deepcopy_dict
    y[deepcopy(key, memo)] = deepcopy(value, memo)
  File "/usr/lib/python2.7/copy.py", line 182, in deepcopy
    rv = reductor(2)
TypeError: 'NoneType' object is not callable

    """
    pdata = PassingData()	
    for colname in rowPointer.table.colnames:
        setattr(pdata, colname, rowPointer[colname])
    return pdata


def castPyTablesEntryIntoPassingData(entry=None):
    """
    2013.3.11 entry is one cell in the array,that is returned from readWhere() query.
        The array is a numpy data structure: array([(1L, '1', '', 2)], 
            dtype=[('id', '<u8'), ('name', '|S512'), ('scientific_name', '|S512'), ('ploidy', '<u2')])
        
    """
    pdata = PassingData()
    for i in range(len(entry.dtype.names)):
        colname = entry.dtype.names[i]
        setattr(pdata, colname, entry[i])
    return pdata


#!/usr/bin/env python3
"""
2009-12-15 http://wj32.wordpress.com/2007/10/08/binary-search-tree-with-heaps-more-features/

	Binary tree (Python)

	Copyright (c) 2007 Wj. All rights reserved.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""

"""
	CHANGELOG:

	Version 1.1:
		* Added changelog.
		* Moved all non-modifying functions to the node class and replaced the code in binary_tree with wrappers.
		* Added a few more list functions and statistical functions.
"""

import math

class node(object):
	"""A node in a tree. Note that all functions which do not modify anything are in this class, while modifying functions are in the binary_tree class."""

	def __init__(self, key = None, data = None):
		self.key = key
		self.data = data
		self.left = None
		self.right = None

	def follow(self, key):
		"""Try to find the specified key in the tree, leaving a "trail" behind of which nodes it has visited. Used to avoid duplicate code."""
		current = self
		trail = []
		value = isinstance(key, node) and key.key or key

		while True:
			if current is None:
				break

			trail.append(current)

			if value == current.key:
				break
			elif value < current.key:
				current = current.left
			elif value > current.key:
				current = current.right
			else:
				# this isn't meant to happen
				break

		return trail

	def find(self, key):
		"""Find a node in the tree."""

		value = isinstance(key, node) and key.key or key
		trail = self.follow(value)

		if trail == []:
			return

		if trail[-1].key == value:
			return trail[-1]

	def depth(self):
		"""Return the how deep the tree is."""

		if not self.left is None:
			leftdepth = self.left.depth()
		else:
			leftdepth = 0

		if not self.right is None:
			rightdepth = self.right.depth()
		else:
			rightdepth = 0

		return max(leftdepth, rightdepth) + 1

	def min(self):
		"""Return the smallest node in the tree."""

		node = self

		while not node.left is None:
			node = node.left

		return node

	def max(self):
		"""Return the biggest node in the tree."""

		node = self

		while not node.right is None:
			node = node.right

		return node

	def dict(self):
		"""Return a dictionary version of the tree"""

		d = {}

		for key in self.listkeys():
			d[key] = self[key]

		return d

	def listrecursive(self):
		"""Return a list of lists of lists (and so on) where each node is represented by [left data right]. So, a simple tree might look like:
			[[None 1 [None 2 None]] 3 [None 4 None]]
		
		The numbers in the example are actually meant to be node objects.
		"""
		l = []

		if not self.left is None:
			l.append(self.left.listrecursive())
		else:
			l.append(None)

		l.append(self)

		if not self.right is None:
			l.append(self.right.listrecursive())
		else:
			l.append(None)

		return l

	def listlayer(self, node, n):
		"""List a layer (used by listlayers())"""

		l = []

		if node is None:
			return None

		if n == 0:
			l.append(node.left)
		else:
			r = self.listlayer(node.left, n - 1)
			l.extend(r is None and [None, None] or r)

		if n == 0:
			l.append(node.right)
		else:
			r = self.listlayer(node.right, n - 1)
			l.extend(r is None and [None, None] or r)

		return l

	def listlayers(self):
		"""Return a list of lists where each list in the list represents a different layer of the tree. An example list of this kind:
			[[3] [1 4] [None 2 None None]
		In this example, 1 is the left child of 3 and 4 is the right child of 3. 2 is the right child of 1.
		"""
		l = []

		l.append([self])

		for i in range(self.depth() - 1):
			l.append(self.listlayer(self, i))

		return l

	def listkeys(self):
		"""Return a list of keys in the tree."""

		l = []

		if not self.left is None:
			l.extend(self.left.listkeys())

		l.append(self.key)

		if not self.right is None:
			l.extend(self.right.listkeys())

		return l

	def listdata(self):
		"""Return a list of data in the tree."""

		l = []

		if not self.left is None:
			l.extend(self.left.listdata())

		l.append(self.data)

		if not self.right is None:
			l.extend(self.right.listdata())

		return l

	def listnodes(self):
		"""Return a list of nodes in the tree."""

		l = []

		if not self.left is None:
			l.extend(self.left.listnodes())

		l.append(self)

		if not self.right is None:
			l.extend(self.right.listnodes())

		return l

	def formattree(self, indentsize = 2, indent = 0, side = "T"):
		"""Return a string with a formatted version of all the nodes in the tree."""

		string = "%s%s:%s: %s\n" % ((indent * " "), side, repr(self.key), repr(self.data))

		if not self.left is None:
			string += self.left.formattree(indentsize, indent + indentsize, "L")

		if not self.right is None:
			string += self.right.formattree(indentsize, indent + indentsize, "R")

		return string

	def formattreemiddle(self, indentsize = 2, indent = 0):
		"""Return a string with a formatted version of all the nodes in the tree."""

		string = ""

		if not self.left is None:
			string += self.left.formattreemiddle(indentsize, indent + indentsize)

		string += "%s%s: %s\n" % ((indent * " "), repr(self.key), repr(self.data))

		if not self.right is None:
			string += self.right.formattreemiddle(indentsize, indent + indentsize)

		return string

	def __getitem__(self, key):
		"""Get an item from the tree. Raises KeyError if the key doesn't exist."""

		item = self.find(key)

		if not item is None:
			return item.data
		else:
			raise KeyError(key)

	def __setitem__(self, key, data):
		"""Set an item. RAISES KeyError IF the key doesn't exist (unlike in binary_tree)."""

		item = self.find(key)

		if item is None:
			raise KeyError(key)
		else:
			item.data = data

	def __contains__(self, key):
		"""Check if a key is present in the tree."""

		item = self.find(key)

		return not item is None

	def __repr__(self):
		return "%s: %s" % (repr(self.key), repr(self.data))

	def __iter__(self):
		"""Generator function"""

		if not self.left is None:
			for node in self.left:
				yield node

		yield self

		if not self.right is None:
			for node in self.right:
				yield node

	def __eq__(self, other):
		"""Compare this node with another node."""

		if not isinstance(other, node):
			return False

		selfkeylist = self.listkeys()
		selfdatalist = self.listdata()
		otherkeylist = other.listkeys()
		otherdatalist = other.listdata()

		return (selfkeylist == otherkeylist) and (selfdatalist == otherdatalist)

	def __ne__(self, other):
		"""Compare this node with another node."""

		return not (self == other)

class binary_tree(object):
	"""A simple binary search tree.

		Nodes can have keys of any type, as long as they can be compared with each other.
	
	Objects can be easily added:
		tree["key"] = "value"
	Or deleted:
		del tree["key"]
	Or tested for membership:
		"key" in tree
	"""

	def __init__(self):
		self.root = None
		self.__count = 0

	def follow(self, key):
		"""Try to find the specified key in the tree, leaving a "trail" behind of which nodes it has visited. Used to avoid duplicate code."""

		if self.root is None:
			return []

		return self.root.follow(key)

	def clear(self):
		"""Delete all items from the tree."""

		self.root = None
		self.__count = 0

	def find(self, key):
		"""Find a node in the tree."""

		if not self.root is None:
			return self.root.find(key)
		else:
			return None

	def insert(self, key, data):
		"""Insert a node into the tree. Raises KeyError if the key already exists."""

		trail = self.follow(key)

		if trail == []:
			self.root = node(key, data)
			self.__count += 1
			return

		if key == trail[-1].key:
			raise KeyError("key %s already in tree" % key)
		elif key < trail[-1].key:
			trail[-1].left = node(key, data)
			self.__count += 1
			return trail[-1].left
		elif key > trail[-1].key:
			trail[-1].right = node(key, data)
			self.__count += 1
			return trail[-1].right

	def delete(self, key):
		"""Remove a node from the tree. Raises KeyError if the key doesn't exist."""

		if self.root is None:
			raise KeyError(key)

		value = isinstance(key, node) and key.key or key
		trail = self.follow(value)
		object = None
		attr = ""

		if value != trail[-1].key:
			raise KeyError(key)

		if len(trail) == 1:
			object = self
			attr = "root"
		elif value < trail[-2].key:
			object = trail[-2]
			attr = "left"
		elif value > trail[-2].key:
			object = trail[-2]
			attr = "right"

		if (trail[-1].left is None) and (trail[-1].right is None):
			setattr(object, attr, None)
		elif trail[-1].left is None:
			setattr(object, attr, trail[-1].right)
		elif trail[-1].right is None:
			setattr(object, attr, trail[-1].left)
		else:
			temp = trail[-1].left

			while not temp.right is None:
				temp = temp.right

			tempdata = temp.data
			tempkey = temp.key
			self.delete(temp)
			trail[-1].data = tempdata
			trail[-1].key = tempkey

		self.__count -= 1

	def depth(self):
		"""Return the how deep the tree is."""

		if not self.root is None:
			return self.root.depth()
		else:
			return 0

	def min(self):
		"""Return the smallest node in the tree."""

		if not self.root is None:
			return self.root.min()
		else:
			return None

	def max(self):
		"""Return the biggest node in the tree."""

		if not self.root is None:
			return self.root.max()
		else:
			return None

	def optimumdepth(self):
		"""Calculate the optimum depth of the tree based on how many nodes there are. The formula is:
			log2(n + 1)
		"""

		return math.log(self.__count + 1, 2)

	def possibleused(self):
		"""Calculate how many nodes could be used based on the depth of the tree. The formula is:
			(2 ^ depth) - 1
		"""

		return (2 ** self.depth()) - 1

	def efficiency(self):
		"""Calculate the efficiency of the tree (how many slots are being wasted). The formula is:
			n / possibleused
		"""

		return float(self.__count) / self.possibleused()

	def dict(self):
		"""Return a dictionary version of the tree."""

		if not self.root is None:
			return self.root.dict()
		else:
			return {}

	def listrecursive(self):
		"""Return a list of lists of lists (and so on) where each node is represented by [left data right]. So, a simple tree might look like:
			[[None 1 [None 2 None]] 3 [None 4 None]]
		
		The numbers in the example are actually meant to be node objects.
		"""

		if not self.root is None:
			return self.root.listrecursive()
		else:
			return []

	def listlayers(self):
		"""Return a list of lists where each list in the list represents a different layer of the tree. An example list of this kind:
			[[3] [1 4] [None 2 None None]
		In this example, 1 is the left child of 3 and 4 is the right child of 3. 2 is the right child of 1.
		"""

		if not self.root is None:
			return self.root.listlayers()
		else:
			return []

	def listkeys(self):
		"""Return a list of keys in the tree. The list will be sorted (this IS a binary search tree...)."""

		if not self.root is None:
			return self.root.listkeys()
		else:
			return []

	def listdata(self):
		"""Return a list of data in the tree. The list will be sorted ACCORDING TO KEY."""

		if not self.root is None:
			return self.root.listdata()
		else:
			return []

	def listnodes(self):
		"""Return a list of nodes in the tree. The list will be sorted ACCORDING TO KEY."""

		if not self.root is None:
			return self.root.listnodes()
		else:
			return []

	def formattree(self, indentsize = 2):
		"""Return a string with a formatted version of all the nodes in the tree."""

		if not self.root is None:
			return self.root.formattree(indentsize)
		else:
			return ""

	def formattreemiddle(self, indentsize = 2):
		"""Return a string with a formatted version of all the nodes in the tree. This version has the root node in the middle."""

		if not self.root is None:
			return self.root.formattreemiddle(indentsize)
		else:
			return ""

	def get(self, key, default=None):
		"""
		2010-1-26
			a dictionary function that was missed before
		"""
		n = self.find(key)
		if n:
			return n.data	# return the value, rather than the node() structure itself.
		return default
	
	def __len__(self):
		"""Return a count of nodes in the tree."""

		return self.__count

	def __getitem__(self, key):
		"""Get an item from the tree. Raises KeyError if the key doesn't exist."""

		if not self.root is None:
			return self.root[key]
		else:
			raise KeyError(key)

	def __setitem__(self, key, data):
		"""Set an item. Adds one if it isn't already present."""

		item = self.find(key)

		if item is None:
			item = self.insert(key, data)
		else:
			item.data = data

	def __delitem__(self, key):
		"""Delete an item from the tree."""

		self.delete(key)

	def __contains__(self, key):
		"""Check if a key is present in the tree."""

		if not self.root is None:
			return key in self.root
		else:
			return False

	def __repr__(self):
		"""Return a string version of the tree."""

		return "binary tree (%s nodes)" % self.__count

	def __iter__(self):
		"""Generator function"""

		if not self.root is None:
			for node in self.root:
				yield node

	def __eq__(self, other):
		"""Compare this tree with another tree."""

		if not isinstance(other, binary_tree):
			return False

		selfkeylist = self.listkeys()
		selfdatalist = self.listdata()
		otherkeylist = other.listkeys()
		otherdatalist = other.listdata()

		return (selfkeylist == otherkeylist) and (selfdatalist == otherdatalist)

	def __ne__(self, other):
		"""Compare this tree with another tree."""

		return not (self == other)

# test program if this file is run
if __name__ == "__main__":
	import random, sys

	tree = binary_tree()
	tree["python"] = None
	tree["c"] = None
	tree["asp"] = None
	tree["ruby"] = None
	tree["java"] = None
	tree["d"] = None
	tree["lisp"] = None
	tree2 = binary_tree()
	tree2["a"] = 1
	tree3 = binary_tree()
	tree3["a"] = 1

	print("Binary Tree Test\n")
	print("Node Count: %d" % len(tree))
	print("Depth: %d" % tree.depth())
	print("Optimum Depth: %f (%d) (%f%% depth efficiency)" % (tree.optimumdepth(), math.ceil(tree.optimumdepth()),
												  math.ceil(tree.optimumdepth()) / tree.depth()))
	print("Min: %s" % repr(tree.min()))
	print("Max: %s" % repr(tree.max()))
	print("Efficiency: %f%% (total possible used: %d, total wasted: %d): " % (tree.efficiency() * 100,
															  len(tree) / tree.efficiency(),
															  (len(tree) / tree.efficiency()) - len(tree)))
	print("List of Layers:\n\t" + repr(tree.listlayers()) + "\n")
	print("\"Recursive\" List:\n\t" + repr(tree.listrecursive()) + "\n")
	print("List of Keys:\n\t" + repr(tree.listkeys()) + "\n")
	print("List of Data:\n\t" + repr(tree.listdata()) + "\n")
	print("List of Nodes:\n\t" + repr(tree.listnodes()) + "\n")
	print("Dictionary:\n\t" + repr(tree.dict()) + "\n")
	print("Formatted Tree:\n" + tree.formattree() + "\n")
	print("Formatted Tree (Root in Middle):\n" + tree.formattreemiddle() + "\n")
	print("tree2 == tree3: " + repr(tree2 == tree3))
	print("\"lisp\" in tree: " + repr("lisp" in tree))
	print("tree[\"d\"]: " + repr(tree["d"]))
	print("tree.root[\"d\"]: " + repr(tree.root["d"]))

	print("Clearing tree...")
	tree.clear()

	count = 1000000
	i = count
	choseni = 0
	chosennumber = 0

	try:
		choseni = random.randrange((count / 2) - (count / 100), (count / 2) + (count / 100))
	except:
		choseni = count / 2

	print("Adding %d random numbers to the tree..." % count)

	while i:
		i -= 1

		n = random.randrange(0, 2147483647)

		try:
			tree[n] = None

			if i == choseni:
				chosennumber = n
		except Exception:
			pass

	print("Done adding, press Enter to continue.")
	sys.stdin.readline()

	print("Finding %d" % chosennumber)
	tree.find(chosennumber)
	print("Found %d" % chosennumber)
	print("Node Count: %d" % len(tree))
	print("Depth: %d" % tree.depth())
	print("Optimum Depth: %f (%d) (%f%% depth efficiency)" % (tree.optimumdepth(), math.ceil(tree.optimumdepth()),
															  math.ceil(tree.optimumdepth()) / tree.depth()))
	print("Min: %s" % repr(tree.min()))
	print("Max: %s" % repr(tree.max()))
	print("Efficiency: %f%% (total possible used: %d, total wasted: %d): " % (tree.efficiency() * 100,
																	  tree.possibleused(),
																	  (tree.possibleused() - len(tree))))
	#print("Formatted Tree (Root in Middle):\n" + tree.formattreemiddle() + "\n")
	print("Removing all nodes...")

	for key in tree.listkeys():
		del tree[key]

	print("Done removing.")



# -*- coding: future_fstrings -*-
"""
2008-01-21
    module for some custom gnome functions
"""
#for Python2&3 compatibility
from __future__ import absolute_import, division, print_function
from builtins import (bytes, str, open, super, range,
    zip, round, input, int, pow, object)

import os, sys
import traceback

def get_chr_pos_from_x_axis_pos(x_axis_pos, chr_gap, chr_id2cumu_size, chr_id_ls):
    """
    2008-02-04
        split out from variation.src.GenomeBrowser.py
    2008-02-03
        get chromosome, position from the x axis position
        chr_id_ls is the sorted version of the keys of chr_id2cumu_size
    """
    chr_id_chosen = None
    position = -1
    for i in range(1, len(chr_id_ls)):
        #the 1st in chr_id_ls is fake chromosome 0
        prev_chr_id = chr_id_ls[i-1]
        chr_id = chr_id_ls[i]
        if chr_id2cumu_size[chr_id]>=x_axis_pos:
            chr_id_chosen = chr_id
            position = x_axis_pos - chr_id2cumu_size[prev_chr_id]
            break
    return chr_id_chosen, position

def foreach_cb(model, path, iter, pathlist):
    """
    Example:
        
        pathlist = []
        self.tree_nodes_selection.selected_foreach(yh_gnome.foreach_cb, pathlist)
        if len(pathlist) >0:
            for i in range(len(pathlist)):
                node_id = self.liststore_nodes[pathlist[i][0]][0]
                self.backend_ins.log_into_node(node_id)
    
    2008-01-21 copied from annot.bin.codense.common
    04-17-05
        used in gui listview, pathfinding.
    """
    pathlist.append(path)	

def edited_cb(cell, path, new_text, user_data):
    """
    2008-02-05
        call back for editable CellRendererText in create_columns()
    """
    liststore, column = user_data
    liststore[path][column] = new_text
    return
    
def create_columns(treeview, label_list, editable_flag_ls=None, liststore=None):
    """
    Examples:
        
        yh_gnome.create_columns(self.treeview1, self.backend_ins.display_job_label_ls)
        self.liststore = gtk.ListStore(*self.backend_ins.display_job_label_type_ls)
        list_2d = self.backend_ins.refresh(username, update_node_info,
            jobs_since, only_running, \
            update_job_info=update_job_info)
        yh_gnome.fill_treeview(self.treeview1, self.liststore, list_2d,
            reorderable=True, multi_selection=True)
    
    2008-02-05
        add editable_flag_ls and liststore
    2008-01-21 copied from annot.bin.codense.common
    04-17-05
        create columns in the treeview in the first refresh
    04-21-05
        remove the old columns and reset the model of treeview
    """
    import gtk
    tvcolumn_dict = {}
    cell_dict = {}
    #remove old columns
    old_column_list = treeview.get_columns()
    for column in old_column_list:
        treeview.remove_column(column)
    treeview.set_model()
        
    for i in range(len(label_list)):
        tvcolumn_dict[i] = gtk.TreeViewColumn(label_list[i])
        # create the TreeViewColumn to display the data
        treeview.append_column(tvcolumn_dict[i])
        # add tvcolumn to treeview
        cell_dict[i] = gtk.CellRendererText()
        # create a CellRendererText to render the data
        if editable_flag_ls and liststore:
            if editable_flag_ls[i]:
                cell_dict[i].set_property('editable', True)
                cell_dict[i].connect('edited', edited_cb, (liststore, i))
        tvcolumn_dict[i].pack_start(cell_dict[i], True)
        # add the cell to the tvcolumn and allow it to expand
        # set the cell "text" attribute to column 0 - retrieve text
        # from that column in liststore
        tvcolumn_dict[i].add_attribute(cell_dict[i], 'text', i)
        tvcolumn_dict[i].set_sort_column_id(i)	# Allow sorting on the column

def fill_treeview(treeview, liststore, list_2d, reorderable=True, multi_selection=True):
    """
    2008-11-07
        put "liststore.append(data)" into try... except ... to handle wrong types in liststore
        fix a bug when list_2d is totally empty
    2008-02-05 add flag multi_selection
    2008-01-21 copied from annot.bin.codense.common
    04-17-05
    05-20-05
        expand the data in list_2d if it's too short
    """
    import gtk
    length_of_treeview = len(treeview.get_columns())
    no_of_rows = len(list_2d)
    for i in range(no_of_rows):
        ls = list_2d[i]
        data = ls[:]
        #copy the list to avoid change the content in ls, 'data=ls' changes the content of ls
        for i in range(length_of_treeview-len(data)):
            data.append('')
        try:
            liststore.append(data)
        except:
            sys.stderr.write("Exception happened for data=%s.\n vs previous data=%s.\n"%\
                (repr(data), repr(list_2d[max(0,i-1)])))
            traceback.print_exc()
            sys.stderr.write('%s.\n'%repr(sys.exc_info()))
    # set the TreeView mode to be liststore
    treeview.set_model(liststore)

    if reorderable:
        if len(list_2d)>0:	#2008-11-07 soemtimes, it's all empty
            for i in range(len(list_2d[0])):
                # make it searchable
                treeview.set_search_column(i)
            
        # Allow drag and drop reordering of rows
        treeview.set_reorderable(True)
    if multi_selection:
        #setting the selection mode
        treeselection = treeview.get_selection()
        treeselection.set_mode(gtk.SELECTION_MULTIPLE)

class Dummy_File:
    """
    2008-01-24
        copied from http://www.daa.com.au/pipermail/pygtk/attachments/20031004/dbb34c38/Py_Shell.py
    """
    def __init__(self, buffer, tag=None):
        """Implements a file-like object for redirect the stream to the buffer"""
        self.buffer = buffer
        self.tag = tag
    
    def write(self, text):
        """Write text into the buffer and apply self.tag"""
        iter=self.buffer.get_end_iter()
        if self.tag:
            self.buffer.insert_with_tags(iter, text, self.tag)
        else:
            self.buffer.insert(iter, text)
    
    def writelines(self, l):
        map(self.write, l)
    
    def flush(self):
        pass
    
    def isatty(self):
        return 1

def subwindow_hide(widget, event, data=None):
    widget.hide()
    return True

#2008-04-21 movied from variation.src.common
#2008-02-04 custom NavigationToolbar
from matplotlib.backends.backend_gtkagg import NavigationToolbar2GTKAgg
class Cursors:  #namespace
    HAND, POINTER, SELECT_REGION, MOVE = range(4)
cursors = Cursors()

class NavigationToolbar2GTKAgg_chromosome(NavigationToolbar2GTKAgg):
    """
    2008-02-04 NavigationToolbar with custom mouse_move()
    """
    def __init__(self, canvas, window):
        NavigationToolbar2GTKAgg.__init__(self, canvas, window)
        self.chr_id2size = {}
        self.chr_id2cumu_size = {}
        self.chr_gap = None
        self.chr_id_ls = None
    
    def update_chr_info(self, chr_id2size, chr_id2cumu_size, chr_gap, chr_id_ls):
        """
        2008-02-04
            fill up the chromosome info
        """
        self.chr_id2size = chr_id2size
        self.chr_id2cumu_size = chr_id2cumu_size
        self.chr_gap = chr_gap
        self.chr_id_ls = chr_id_ls
    
    def mouse_move(self, event):
        """
        2008-02-04
            custom the label shown for 'motion_notify_event'
            self._idDrag=self.canvas.mpl_connect('motion_notify_event', self.mouse_move)
        """
        #print 'mouse_move', event.button

        if not event.inaxes or not self._active:
            if self._lastCursor != cursors.POINTER:
                self.set_cursor(cursors.POINTER)
                self._lastCursor = cursors.POINTER
        else:
            if self._active=='ZOOM':
                if self._lastCursor != cursors.SELECT_REGION:
                    self.set_cursor(cursors.SELECT_REGION)
                    self._lastCursor = cursors.SELECT_REGION
                if self._xypress:
                    x, y = event.x, event.y
                    lastx, lasty, a, ind, lim, trans= self._xypress[0]
                    self.draw_rubberband(event, x, y, lastx, lasty)
            elif (self._active=='PAN' and
                  self._lastCursor != cursors.MOVE):
                self.set_cursor(cursors.MOVE)

                self._lastCursor = cursors.MOVE

        if event.inaxes and event.inaxes.get_navigate():

            try:
                #2008-02-04 custom here
                if self.chr_id2size and self.chr_id2cumu_size and \
                    self.chr_gap!=None and self.chr_id_ls:
                    x_axis_pos_int = int(round(event.xdata))
                    chr_id, position = get_chr_pos_from_x_axis_pos(
                        x_axis_pos_int, self.chr_gap,
                        self.chr_id2cumu_size, self.chr_id_ls)
                    chr_perc = -1	#default
                    if chr_id in self.chr_id2size:
                        chr_perc = position/float(self.chr_id2size[chr_id])*100
                        if chr_perc>100 or chr_perc<0:	#out of range
                            chr_perc = -1
                    if position==-1:	#not within any chromosome
                        position = x_axis_pos_int
                    s = 'chr=%s, pos=%i(%.1f%%), y=%s'%(chr_id, position,
                        chr_perc, event.ydata)
                else:
                    s = event.inaxes.format_coord(event.xdata, event.ydata)
            except ValueError: pass
            except OverflowError: pass
            else:
                if len(self.mode):
                    self.set_message('%s : %s' % (self.mode, s))
                else:
                    self.set_message(s)
        else: self.set_message(self.mode)

def getDataOutOfTextEntry(widget, data_type=None, filter_func=None, default=None):
    """
    2010-6-14
        add argument default
    2010-4-27
        This function fetches the data present in a text widget.
        
        widget could be text_entry, combobox, ComboBoxEntry
        
        Example for filter_func
            filter_func = lambda x: x.split('_')[0]
    """
    if hasattr(widget, 'get_text'):
        text = widget.get_text()
    elif hasattr(widget, 'get_active_text'):	# combobox or ComboBoxEntry
        text = widget.get_active_text()
    else:
        text = ''
    if text and filter_func:
        text = filter_func(text)
    if data_type:
        if text:
            text = data_type(text)
        else:
            text = None
    if default is not None and text is None:
        text = default
    return text


#!/usr/bin/env python3
"""
Description:
	20130731
	abstract walker class for genome files
"""

import sys, os, math
import csv, random, numpy
from palos import ProcessOptions, getListOutOfStr, PassingData, utils
from palos.db import GenomeDB
from palos.io.AbstractMatrixFileWalker import AbstractMatrixFileWalker

ParentClass = AbstractMatrixFileWalker

class AbstractGenomeFileWalker(ParentClass):
	__doc__ = __doc__
	option_default_dict = ParentClass.option_default_dict.copy()
	option_default_dict.update(ParentClass.genome_db_option_dict.copy())
	#option_default_dict.update(ParentClass.db_option_dict.copy())
	
	genome_option_dict = {
		('chromosomeHeader', 1, ): ['chromosome', '', 1, 'header of the column that designates chromosome info' ],\
		('positionHeader', 1, ): ['position', '', 1, 'header of the column that designates position of each locus on chromosome' ],\
		('tax_id', 0, int): [3702, '', 1, 'taxonomy ID of the organism from which '
			'to retrieve the chromosome info', ],\
		('sequence_type_id', 0, int): [1, '', 1, "sequence_type_id (annot_assembly) of the chromosomes, "
			"to retrieve the chromosome info. 1: assembledChromosome, 9: Scaffolds", ],\
		('chrOrder', 0, int): [1, '', 1, 'how to order chromosomes. '
			'1: column genome_order in db; 2: by chromosome size, descending', ],\
		}
	option_default_dict.update(genome_option_dict)
	option_default_dict.update({
					
					
					})
	def __init__(self, inputFnameLs=None, **keywords):
		"""
		"""
		ParentClass.__init__(self, inputFnameLs=inputFnameLs, **keywords)	#self.connectDB() called within its __init__()
	
	def _loadGenomeStructureFromDB(self):
		"""
		2013.07.31
			so that both AbstractGenomeFileWalker and PlotGenomeWideData could call it
		"""
		db_genome = GenomeDB.GenomeDatabase(drivername=self.genome_drivername, username=self.genome_db_user,
						db_passwd=self.genome_db_passwd, hostname=self.genome_hostname, dbname=self.genome_dbname, \
						schema=self.genome_schema)
		db_genome.setup(create_tables=False)
		self.db_genome = db_genome
		
		#chrOrder=1 is to order chromosomes alphabetically
		self.oneGenomeData = db_genome.getOneGenomeData(tax_id=self.tax_id, chr_gap=0, chrOrder=self.chrOrder, \
													sequence_type_id=self.sequence_type_id)
		self.chr_id2cumu_start = self.oneGenomeData.chr_id2cumu_start
		#2013.2.18 get centromere locations
		self.chr_id2centromere = self.oneGenomeData.chr_id2centromere
		
	
	def setup(self, **keywords):
		"""
		2013.07.31
			construct an RBTree dictionary map between windows and their data
		"""
		ParentClass.setup(self, **keywords)
		self._loadGenomeStructureFromDB()
	

if __name__ == '__main__':
	main_class = AbstractGenomeFileWalker
	po = ProcessOptions(sys.argv, main_class.option_default_dict, error_doc=main_class.__doc__)
	instance = main_class(po.arguments, **po.long_option2value)
	instance.run()


# -*- coding: future_fstrings -*-
import os, sys
from . utils import getListOutOfStr

def process_function_arguments(keywords, argument_default_dict, error_doc='',
    class_to_have_attr=None, howto_deal_with_required_none=1,
    default_value_in_list=0):
    """
    2008-04-28
        add argument default_value_in_list to differentiate 
            argument_default_dict whose value is a list which potentially
        could include option instructions (short_option, has_argument,
            description_for_option).
    2008-04-10
        add argument: howto_deal_with_required_none.
        if =1, prompt user to enter. else, terminate the program.
    2008-04-09
        if required argument is not given, prompt the user to get it.
    2008-04-02
        add class_to_have_attr to assign the argument and values
    2008-02-28
        example of argument_default_dict:
    
        argument_default_dict = {
            ('hostname',1, ):'papaya.usc.edu',\
            ('dbname',1, ):'stock_250k',\
            ('user',1, ):None,\
            ('passwd',1, ):None,\
            ('experimenter',0,):None,\
            ('mapping_file',1,):None,\
            ('input_dir',1, ):None,\
            ('output_dir',1, ):'/Network/Data/250k/db/raw_data/',\
            ('array_data_table',0, ):'array_data',\
            ('probes_table',0, ):'probes',\
            ('strain_info_table',0, ):'strain_info',\
            ('array_info_table',1, ):'array_info',\
            ('commit',0, int):0,\
            ('debug',0, int):0,\
            ('report',0, int):0}

    process arguments from keywords based on argument_default_dict
    
    keywords is a dictionary of given arguments. 'argument_name':'argument_value'
    
    argument_default_dict is a dictionary of default arguments.
        the key is a tuple: ('argument_name', is_argument_required, argument_type)
            and argument_type is optional.
        the value is could just be the default_value or a list = [default_value,
            'short_option', 'long_option', has_argument, description_for_option]
            'short_option', has_argument, description_for_option are all optional.
    """
    import sys
    ad = {}
    argument_key_ls = sorted(argument_default_dict)
    #to keep 'user' appearing in front of 'password'.
    argument_key_ls.reverse()
    for argument_key in argument_key_ls:
        if default_value_in_list:
            default_value = argument_default_dict[argument_key][0]
        else:
            default_value = argument_default_dict[argument_key]
        argument, is_argument_required = argument_key[0:2]
        if len(argument_key)>2:
            argument_type = argument_key[2]
        else:
            argument_type = None
        if keywords.has_key(argument):
            if keywords[argument]!='' and keywords[argument]!=None:
                #only when keywords has this argument and it's not nothing.
                #  change default_value
                if argument_type!=None:	#cast to the desired type
                    default_value = argument_type(keywords[argument])
                else:
                    default_value = keywords[argument]
        if is_argument_required==1 and (default_value=='' or default_value==None):
            if howto_deal_with_required_none==1:
                if argument=='passwd' or argument=='Passwd' or argument=='password':
                    import getpass
                    default_value = getpass.getpass(f"{argument}: ")
                else:
                    default_value = input(f"{argument}: ")
            else:
                if error_doc:
                    sys.stderr.write(error_doc)
                sys.stderr.write('Error: %s is required but %s given.\n'%(\
                    argument, default_value))
                sys.exit(5)
        if argument_type!=None:	#cast to the desired type
            default_value = argument_type(default_value)
        if class_to_have_attr:
            setattr(class_to_have_attr, argument, default_value)
        ad[argument] = default_value
    return ad

def process_options(argv_list, option_default_dict, error_doc=''):
    """
    2008-04-25
        add option 'help' into option_default_dict
        program exits if '-h' or '--help' is set.
        throw away 'try ... except ...' around getopt
    2008-04-20
        wraps the option handling (getopt), the usual block 
            underneath "if __name__ == '__main__':".
    example of option_default_dict:
    
        option_default_dict = {
            ('z', 'hostname', 1, 'hostname of the db server', 1, ): 'papaya.usc.edu',\
            ('d', 'dbname', 1, '', 1, ): 'stock_250k',\
            ('u', 'user', 1, '', 1, ):None,\
            ('p', 'passwd', 1, '', 1, ):None,\
            ('t', 'call_info_table', 1, '', 1, ): 'call_info',\
            ('i', 'cmp_data_filename', 1, 'the data file to be compared with.', 1, ): None,\
            ('q', 'call_QC_table', 1, '', 1, ): 'call_QC',\
            ('m', 'QC_method_id', 1, 'id in table QC_method', 1, int): None,\
            ('c', 'commit', 0, 'commit db transaction', 0, int):0,\
            ('b', 'debug', 0, 'toggle debug mode', 0, int):0,\
            ('r', 'report', 0, 'toggle report, more verbose stdout/stderr.', 0, int):0}
    
        option_default_dict is a dictionary for option handling, including
             argument_default_dict info.
        the key is a tuple, ('short_option', 'long_option', has_argument, 
            description_for_option, is_option_required, argument_type)
        argument_type is optional
    """
    import sys
    if len(argv_list) == 1:
        print(error_doc)
        sys.exit(6)
    option_default_dict[('h', 'help', 0, 'Display this documentation', 0, int)] =0
    
    #prepare long_options_list and short_options_str for getopt
    long_options_list = []
    short_options_list = []
    short_option2long_option = {}
    long_option2has_argument = {}
    for option_key in option_default_dict:
        short_option, long_option, has_argument = option_key[0:3]
        if has_argument:
            long_options_list.append('%s='%long_option)
        else:
            long_options_list.append('%s'%long_option)
        if has_argument:
            short_options_list.append('%s:'%short_option)
        else:
            short_options_list.append(short_option)
        short_option2long_option[short_option] = long_option
        long_option2has_argument[long_option] = has_argument
    short_options_str  = ''.join(short_options_list)
    
    #handle options
    import getopt, traceback
    opts, args = getopt.getopt(argv_list[1:], short_options_str, long_options_list)
    opts_dict = {}
    for opt, arg in opts:
        if opt[1]=='-':
            #it's long option
            long_option = opt[2:]
        else:
            #it's short option
            short_option = opt[1:]
            long_option = short_option2long_option[short_option]
        if long_option=='help':
            print(error_doc)
            sys.exit(0)
        if long_option2has_argument[long_option]:
            opts_dict[long_option] = arg
        else:
            #toggle the bit for options which don't have arguments
            opts_dict[long_option] = 1
    """
    except:
        traceback.print_exc()
        print sys.exc_info()
        print(error_doc)
        sys.exit(2)
    """
    return opts_dict


class ProcessOptions:
    """
    2008-04-28
        a class to replace process_options(), process_function_arguments(), 
        generate_program_doc(), turn_option_default_dict2argument_default_dict()
    """
    getopt_ready = 0
    def __init__(self, argv_list=None, option_default_dict=None, error_doc='',
        program_name=None):
        self.argv_list = argv_list
        self.option_default_dict = option_default_dict
        self.error_doc = error_doc
        if program_name!=None and program_name!='':
            self.program_name = program_name
        elif (program_name==None or program_name=='') and type(argv_list)==list\
                and len(argv_list)>0:
            self.program_name = argv_list[0]
        else:
            self.program_name = ''
        #2011-7-13	to store all the remaining arguments
        self.args = []
    
    def prepare_for_getopt(self, option_default_dict):
        """
        2012.8.17
            bugfix. for the options that do not have short-option set before hand,
             the bug assumed all of them need arguments.
        2008-09-09
            in the case that short_option is not provided,
                get the has_argument and description_for_option if they are provided.
        2008-05-02
            correct a bug in deal with the rest with no short_option specified
        2008-04-28
            2nd-generation of process_options()
            option_default_dict becomes an enlargement of and compatible to
                argument_default_dict.
            
        option_default_dict = {
            ('hostname', 1, ): ['papaya.usc.edu', 'z', 1, 'hostname of the db server',],\
            ('dbname', 1, ): ['stock_250k', 'd', 1, '', ],\
            ('user', 1, ): [None, 'u', 1, ],\
            ('passwd', 1, ): [None, 'p', ],\
            ('commit', 0, int): [0, 'c', 0, 'commit db transaction', ],\
            ('debug', 0, int): [0, 'b', 0, 'toggle debug mode', ],\
            ('report', 0, int):[0, 'r', 0, \
                'toggle report, more verbose stdout/stderr.', ]}
    
        option_default_dict is a dictionary for option handling,
                    including argument_default_dict info.
            the key is a tuple, ('argument_name', is_argument_required,
                argument_type) and argument_type is optional.
            
            the value could just be the default_value or a list =[
                default_value, 'short_option', has_argument, description_for_option]
            'short_option', has_argument, description_for_option are all
                orderly optional. You can just supply 'short_option' or
                'short_option', has_argument.
        """
        import sys
        option_default_dict[('help', 0, int)] = [0, 'h', 0, \
            'Display this documentation', ]
        #for generate_program_doc()
        std_option_default_dict = option_default_dict.copy()
        
        #prepare long_options_list and short_options_str for getopt
        long_options_list = []
        short_options_list = []
        short_option2long_option = {}
        long_option2has_argument = {}
        options_with_no_short_option = []
        for option_key, option_value in option_default_dict.items():
            long_option = option_key[0]
            if type(option_value)==list and len(option_value)>=4:
                default_value, short_option, has_argument, \
                    description_for_option = option_value[:4]
            elif type(option_value)==list and len(option_value)>=3:
                default_value, short_option, has_argument = option_value[:3]
                description_for_option = ''
            elif type(option_value)==list and len(option_value)==2:
                default_value, short_option = option_value[:2]
                has_argument = 1
                #not specified, assuming this option has argument.
                description_for_option = ''
            else:
                #04/28/08 deal with these later. use short options that are not
                #  used by others
                options_with_no_short_option.append(option_key)
                continue
            
            if short_option==None or short_option=='':
                #2012.8.17 handle these no-short-option options later.
                options_with_no_short_option.append(option_key)
                continue
            #prepare short options
            if short_option in short_option2long_option:
                sys.stderr.write("Error: short option %s already used by %s.\n"%\
                    (short_option, short_option2long_option[short_option]))
                sys.exit(3)
            short_option2long_option[short_option] = long_option
            if has_argument:
                short_options_list.append('%s:'%short_option)
            else:
                short_options_list.append(short_option)
            
            #prepare long options
            if long_option in long_option2has_argument:
                sys.stderr.write("Error: long option %s already used.\n"%(long_option))
                sys.exit(3)
            long_option2has_argument[long_option] = has_argument
            if has_argument:
                long_options_list.append('%s='%long_option)
            else:
                long_options_list.append('%s'%long_option)
            
            std_option_default_dict[option_key] = [default_value, short_option,\
                has_argument, description_for_option]
        #deal with the rest with no short_option specified
        #always assume has_argument=1
        chr_ordinal_ls = list(range(97, 97+26)) + list(range(65, 65+26))
        #potential candidates
        eng_letter_ls = map(chr, chr_ordinal_ls)
        eng_letters = ''.join(eng_letter_ls)
        for option_key in options_with_no_short_option:
            long_option = option_key[0]
            option_value = option_default_dict[option_key]
            #2008-09-09 get the has_argument and description_for_option if they are provided.
            if type(option_value)==list and len(option_value)>=4:
                has_argument = option_value[2]
                description_for_option = option_value[3]
            elif type(option_value)==list and len(option_value)>=3:
                has_argument = option_value[2]
                description_for_option = ''
            else:
                has_argument = 1
                description_for_option = ''
            short_option = None
            for letter in long_option + eng_letters:
                #start with the long_option, then try all english letters
                if letter not in eng_letter_ls:
                    #this character has to be english letters.
                    # long_option might contain non-letters. like '_'.
                    continue
                if letter not in short_option2long_option:
                    short_option2long_option[letter] = long_option
                    short_options_list.append('%s:'%letter)
                    short_option = letter
                    break
            if long_option in long_option2has_argument:
                sys.stderr.write("Error: long option %s already used.\n"%\
                    (long_option))
                sys.exit(3)
            long_option2has_argument[long_option] = has_argument
            if has_argument:
                #2012.8.17 bugfix. whether to add the = depends on has_argument
                long_options_list.append('%s='%long_option)
            else:
                long_options_list.append('%s'%long_option)
            #long_options_list.append('%s='%long_option)
            std_option_default_dict[option_key] = [option_value[0], \
                short_option, has_argument, description_for_option]
            
        self.short_options_str  = ''.join(short_options_list)
        self.short_option2long_option = short_option2long_option
        self.long_option2has_argument = long_option2has_argument
        self.long_options_list = long_options_list
        self.std_option_default_dict = std_option_default_dict
        self.getopt_ready = 1
    
    @property
    def program_doc(self):
        """
        2008-05-02
            fix a bug in toggle need_star
        2008-04-28
            sort the options in short_option ascending order.
            structure of std_option_default_dict changed.
        2008-04-28
            2nd-generation of generate_program_doc()
            option_default_dict becomes an enlargement of and compatible to
                argument_default_dict.
        """
        if self.getopt_ready==0:
            self.prepare_for_getopt(self.option_default_dict)
        
        usage_str = 'Usage: %s [OPTIONS]'%self.program_name
        argument_list_str_ls = ['Argument List:']
        
        #sort the options in short_option ascending order
        option_key_ls_to_be_sorted = []
        for option_key, option_value in self.std_option_default_dict.items():
            if option_key[0] == 'help':	#help would be appended in the end
                continue
            option_key_ls_to_be_sorted.append((option_value[1], option_key))
            #option_value[1] = short_option
        #print(option_key_ls_to_be_sorted)	
        #option_key_ls_to_be_sorted.sort()
        # #20191210 error in python3: '<' not supported
        #  between instances of 'NoneType' and 'str'
        
        for short_option, option_key in option_key_ls_to_be_sorted:
            option_value = self.std_option_default_dict[option_key]
            default_value, short_option, has_argument, description_for_option =\
                option_value
            long_option, is_option_required = option_key[0:2]
            need_star = 0	#required and default_value is empty
            this_argument_ls = ['    ']
            if is_option_required and (default_value==None or default_value==''):
                usage_str += ' --%s %s'%(long_option, long_option.upper())
                need_star = 1
            
            if has_argument:
                if short_option:
                    this_argument_ls.append(f'-{short_option} ...,  ')
                    this_argument_ls.append(f'--{long_option}=...')
                else:
                    this_argument_ls.append(f'--{long_option}=...')
            else:
                if short_option:
                    this_argument_ls.append(f'-{short_option}  --{long_option}')
                else:
                    this_argument_ls.append(f'--{long_option}')
            
            if need_star:
                this_argument_ls.append('*')
            if description_for_option:
                if description_for_option[-1]=='.':
                    this_argument_ls.append('    %s '%description_for_option)
                else:
                    this_argument_ls.append('    %s. '%description_for_option)
            else:
                this_argument_ls.append('    ')
            if has_argument and default_value!=None and default_value!='':
                this_argument_ls.append('"%s"(default)'%default_value)
            argument_list_str_ls.append(''.join(this_argument_ls))
        this_argument_ls = ['    ', '-h,  --help', '  ', 'Display this documentation']
        argument_list_str_ls.append(''.join(this_argument_ls))
        argument_list_str_ls.append(
            "\nFor required(*) options, if no argument is given, you'll be prompted.")
        program_doc = '\n%s\n\n'%usage_str
        program_doc += '\n'.join(argument_list_str_ls)
        program_doc += '\n'
        return program_doc
    
    @property
    def long_option2value(self):
        """
        """
        if self.getopt_ready==0:
            self.prepare_for_getopt(self.option_default_dict)
        
        error_doc = self.program_doc + self.error_doc
        
        import sys
        if len(self.argv_list) == 1:
            print(error_doc)
            sys.exit(3)
        
        import getopt, traceback
        opts, args = getopt.getopt(self.argv_list[1:], self.short_options_str,\
            self.long_options_list)
        self.args = args
        opts_dict = {}
        for opt, arg in opts:
            if opt[1]=='-':
                #it's long option
                long_option = opt[2:]
            else:
                #it's short option
                short_option = opt[1:]
                long_option = self.short_option2long_option[short_option]
            if long_option=='help':
                print(error_doc)
                sys.exit(0)
            if self.long_option2has_argument[long_option]:
                opts_dict[long_option] = arg
            else:
                #toggle the bit for options which don't have arguments
                opts_dict[long_option] = 1
        """
        except:
            traceback.print_exc()
            print sys.exc_info()
            print(error_doc)
            sys.exit(2)
        """
        return opts_dict
    
    @property
    def arguments(self):
        """
        2011-7-14
            to return all the remaining un-parsed arguments
        """
        self.opts_dict = self.long_option2value
        return self.args
    
    @classmethod
    def process_function_arguments(cls, keywords, argument_default_dict,
        error_doc='', class_to_have_attr=None, howto_deal_with_required_none=1):
        """
        2009-10-07
            "is not ''" works for character numpy.array, while "!=''" doesn't
                in if-condition.
        2008-10-25 
            if default_value is None, and no value is given by a user,
                no casting into type specified.
        2008-07-09
            not just ='passwd', if the option name (variable 'argument' here)
                contains 'passwd', use getpass().
        2008-05-06
            return ad as well, just like the standalone function
        2008-04-28
            largely copied from the standalone process_function_arguments()
            argument_default_dict = option_default_dict in prepare_for_getopt()
            The only difference is that value of this argument_default_dict
                could be non-list, which is plain default_value.
            To be compatible with previous argument_default_dict.
        """
        argument_key_ls = sorted(argument_default_dict, reverse=True)
        #to keep 'user' appearing in front of 'password'.
        ad = {}
        for argument_key in argument_key_ls:
            if type(argument_default_dict[argument_key])!=list:
                default_value = argument_default_dict[argument_key]
            else:
                default_value = argument_default_dict[argument_key][0]
            argument, is_argument_required = argument_key[0:2]
            if len(argument_key)>2:
                argument_type = argument_key[2]
            else:
                argument_type = None
            if argument in keywords:
                if keywords[argument] is not '' and \
                    keywords[argument] is not None:
                    # 2009-10-07 "is not ''" works for character numpy.array,
                    #   while "!=''" doesn't in if-condition.
                    #only when keywords has this argument and it's not nothing.
                    #   change default_value
                    if argument_type!=None:	#cast to the desired type
                        default_value = argument_type(keywords[argument])
                    else:
                        default_value = keywords[argument]
            if is_argument_required==1 and (default_value=='' or default_value==None):
                if howto_deal_with_required_none==1:
                    if argument.find('passwd')!=-1 or \
                        argument.find('Passwd')!=-1 or argument.find('password')!=-1:
                        import getpass
                        default_value = getpass.getpass(f"{argument}:")
                    else:
                        default_value = input(f"{argument}:")
                else:
                    if error_doc:
                        sys.stderr.write(error_doc)
                    sys.stderr.write('Error: %s is required but %s given.\n'%\
                        (argument, default_value))
                    sys.exit(4)
            if argument_type!=None and default_value is not None:
                #cast to the desired type 2008-10-25 default_value is not None
                default_value = argument_type(default_value)
            if class_to_have_attr is not None:
                setattr(class_to_have_attr, argument, default_value)
            ad[argument] = default_value
        return ad
    
    @classmethod
    def processListArguments(cls, listArgumentName_data_type_ls=None,
        emptyContent=[], class_to_have_attr=None):
        """
        2013.3.22 skip getListOutOfStr() if listArgumentValue is
             already a list or tuple. some conversion done already
        2012.10.5 copied from AbstractAlignmentAndVCFWorkflow.py
        2012.8.15
        i.e.:
        listArgumentName_data_type_ls = [(
            'ind_seq_id_ls', int), ("ind_aln_id_ls", int), \
            ("site_id_ls", int), ('country_id_ls', int), ('tax_id_ls', int)]
        listArgumentName2hasContent = ProcessOptions.processListArguments(
            listArgumentName_data_type_ls, emptyContent=[])
        """
        listArgumentName2hasContent = {}
        for listArgumentName, data_type in listArgumentName_data_type_ls:
            listArgumentValue = getattr(class_to_have_attr, listArgumentName, None)
            if listArgumentValue:
                if isinstance(listArgumentValue, list) or \
                    isinstance(listArgumentValue, tuple):
                    pass
                    #already a list or tuple. some conversion done already
                else:
                    listArgumentValue = getListOutOfStr(listArgumentValue, \
                        data_type=data_type)
                setattr(class_to_have_attr, listArgumentName, listArgumentValue)
                listArgumentName2hasContent[listArgumentName]=True
            else:
                setattr(class_to_have_attr, listArgumentName, emptyContent)
                listArgumentName2hasContent[listArgumentName]=False
        return listArgumentName2hasContent
    
    @classmethod
    def removeCertainOptions(cls, option_default_dict=None,
        option_dict_to_remove=None):
        """
        2012.11.20
            convenient function for removing set of options
                (option_dict_to_remove).
        """
        for option_key in option_dict_to_remove:
            if option_key in option_default_dict:
                option_default_dict.pop(option_key)
        
    
def turn_option_default_dict2argument_default_dict(option_default_dict):
    """
    2008-04-20
        option_default_dict contains the info of argument_default_dict.
        to avoid repetitive code
    """
    argument_default_dict = {}
    for option_key, default_value in option_default_dict.items():
        short_option, long_option, has_argument, description_for_option,\
            is_option_required = option_key[0:5]
        if len(option_key)==6:
            argument_type = option_key[5]
            argument_key = (long_option, is_option_required, argument_type)
        else:
            argument_key = (long_option, is_option_required, )
        argument_default_dict[argument_key] = default_value
    return argument_default_dict

def generate_program_doc(program_name, option_default_dict):
    """
    2008-04-25
        cosmetic change. if description_for_option already has . in the end.
        No '.' appended.
    2008-04-25
        add -h, --help to program_doc
    2008-04-24
        generates better doc
    2008-04-21
        automatically generate documentation like
             "Usage: ... \n Argument List: ... " from option_default_dict
    
    example of option_default_dict:
    
    option_default_dict = {
        ('z', 'hostname', 1, 'hostname of the db server', 1, ): 'papaya.usc.edu',\
        ('d', 'dbname', 1, '', 1, ): 'stock_250k',\
        ('u', 'user', 1, '', 1, ):None,\
        ('p', 'passwd', 1, '', 1, ):None,\
        ('t', 'call_info_table', 1, '', 1, ): 'call_info',\
        ('i', 'cmp_data_filename', 1, 'the data file to be compared with.', 1, ): None,\
        ('q', 'call_QC_table', 1, '', 1, ): 'call_QC',\
        ('m', 'QC_method_id', 1, 'id in table QC_method', 1, int): None,\
        ('c', 'commit', 0, 'commit db transaction', 0, int):0,\
        ('b', 'debug', 0, 'toggle debug mode', 0, int):0,\
        ('r', 'report', 0, 'toggle report, more verbose stdout/stderr.', 0, int):0}
    
        option_default_dict is a dictionary for option handling,
             including argument_default_dict info.
        the key is a tuple, ('short_option', 'long_option', has_argument,
            description_for_option, is_option_required, argument_type)
        argument_type is optional
    """
    usage_str = 'Usage: %s [OPTIONS]'%program_name
    argument_list_str_ls = ['Argument List:']
    option_key_ls = sorted(option_default_dict)
    for option_key in option_key_ls:
        default_value = option_default_dict[option_key]
        short_option, long_option, has_argument, description_for_option, \
            is_option_required = option_key[0:5]
        need_star = 0
        this_argument_ls = ['\t']
        if is_option_required and default_value==None or default_value=='':
            usage_str += ' -%s %s'%(short_option, long_option.upper())
            need_star = 1
        if has_argument:
            this_argument_ls.append('-%s ...,\t--%s=...'%(short_option, \
                long_option))
        else:
            this_argument_ls.append('-%s,\t--%s'%(short_option, long_option))
        if need_star:
            this_argument_ls.append('*')
        if description_for_option:
            if description_for_option[-1]=='.':
                this_argument_ls.append('\t%s '%description_for_option)
            else:
                this_argument_ls.append('\t%s. '%description_for_option)
        else:
            this_argument_ls.append('\t')
        if has_argument and default_value!=None and default_value!='':
            this_argument_ls.append('"%s"(default)'%default_value)
        argument_list_str_ls.append(''.join(this_argument_ls))
    this_argument_ls = ['\t', '-h,\t--help', '\tDisplay this documentation']
    argument_list_str_ls.append(''.join(this_argument_ls))
    argument_list_str_ls.append(
        "\nFor required(*) options, if no argument is given, you'll be prompted.")
    program_doc = '\n%s\n\n'%usage_str
    program_doc += '\n'.join(argument_list_str_ls)
    program_doc += '\n'
    return program_doc




#!/usr/bin/env python3
from numpy import abs, array, average, corrcoef, mat, shape, std, sum, transpose, zeros
from numpy.linalg import svd

try:
	import c_nipals
	import_ok = True
except:
	import_ok = False
	 
__author__ = "Henning Risvik"
__date__ = "February 2008"
__version__ = "1.1.01"
	
################### Preprocessing Methods ###################
def mean_center(X):
	"""
	
	@param X: 2-dimensional matrix of number data 
	@type X: numpy array
	
	
	@return: Mean centered X (always has same dimensions as X)
	
	"""
	(rows, cols) = shape(X)
	new_X = zeros((rows, cols), float)
	_averages = average(X, 0)
		
	for row in range(rows):
		new_X[row, 0:cols] = X[row, 0:cols] - _averages[0:cols]
	return new_X
		
		
def standardization(X):        
	"""
	
	@param X: 2-dimensional matrix of number data 
	@type X: numpy array
	
	
	@return: Standardized X (always has same dimensions as X)
	
	"""
	(rows, cols) = shape(X)
	new_X = zeros((rows, cols))
	_STDs = std(X, 0)
		
	for value in _STDs:
		if value == 0: raise ZeroDivisionError('division by zero, cannot proceed')
		
	for row in range(rows):
		new_X[row, 0:cols] = X[row, 0:cols] / _STDs[0:cols]
	return new_X        


	   
################### NIPALS array help functions ###################
def get_column(E):
	"""
	@param E: 2-dimensional matrix of number data 
	@type E: numpy array    
	
	@return: a column which has a non-zero eigenvalue
	"""
	(rows, cols) = shape(E)
	for col_ind in range(cols):
		t = E[:,col_ind] # extract a column
		if (eigenvalue_vec(t) > 0):
			return t
	raise ValueError('all column vectors in E have zero-eigenvalues') # error: sum of matrix is 0


def get_column_mat(E):
	"""
	@param E: 2-dimensional matrix of number data 
	@type E: numpy matrix    
	
	@return: a column which has a non-zero eigenvalue
	"""
	(rows, cols) = shape(E)
	for col_ind in range(cols):
		t = E[:,col_ind] # extract a column
		eig = transpose(t)*t
		if (eig > 0):
			return t
	raise ValueError('all column vectors in E have zero-eigenvalues') # error: sum of matrix is 0

def eigenvalue_vec(v):
	"""
	@param v: Vector of number data.
	@type v: numpy array     
	
	@return: (transpose(v) * v) the eigen-value of "matrix" v as float or int
	"""
	return sum(v * v);

def mat_prod(A, x):
	"""    
	@param A: 2-dimensional matrix of number data.
	@type A: numpy array 
	
	@param x: Vector of number data.
	@type x: numpy array         

	@return: b of (Ax = b). Product of:  matrix A (m,n) * vector x (n) = vector b (m)
	"""
	#m = A.shape[0]
	#b = zeros((m), float)
	
	# calc: Ax = b
	#for i in range(m):
	#    b[i] = sum(A[i,:]*x)

	return array(map(lambda a: sum(a[:]*x), A))

def remove_tp_prod(E, t, p):
	"""
	
	sets: E = E - (t*transpose(p))
	
	E: (m, n)-matrix, (t*transpose(p)): (m, n)-matrix   
	
	
	@param E: 2-dimensional matrix of number data.
	@type E: numpy array 
	
	@param t: Vector of number data. Current Scores (of PC_i).
	@type t: numpy array         

	@param p: Vector of number data. Current Loading (of PC_i).
	@type p: numpy array   


	@return: void    


	"""
	
	m = E.shape[0]
	for i in range(m):
		E[i, :] = E[i, :] - (t[i] * p)
	   

################### NIPALS Algorithm ###################
"""
  Estimation of PC components with the iterative NIPALS method: 


  E[0] = mean_center(X)  (the E-matrix for the zero-th PC)

  t = E(:, 0)  (a column in X (mean centered) is set as starting t vector)

  for i=1 to (PCs):
	
	1  p=(E[i-1]'t) / (t't)  Project X onto t to find the corresponding loading p
	
	2  p = p * (p'p)^-0.5  Normalise loading vector p to length 1
	
	3  t = (E[i-1]p) / (p'p)  Project X onto p to find corresponding score vector t
	
	4  Check for convergence, if difference between eigenval_new and eigenval_old is larger than threshold*eigenval_new return to step 1
	
	5  E[i] = E[i-1] - tp'  Remove the estimated PC component from E[i-1] 
	
"""    
def nipals_mat(X, PCs, threshold, E_matrices):
	"""
	
	
	PCA by NIPALS using numpy matrix
	
	
	@param X: 2-dimensional matrix of number data. 
	@type X: numpy array
	
	@param PCs: Number of Principal Components.
	@type PCs: int
	
	@param threshold: Convergence check value. For checking on convergence to zero (e.g. 0.000001). 
	@type threshold: float
	
	@param E_matrices: If E-matrices should be retrieved or not. E-matrices (for each PC) or explained_var (explained variance for each PC).
	@type E_matrices: bool
	
	@return: (Scores, Loadings, E)

	"""
	(rows, cols) = shape(X)

	maxPCs = min(rows, cols) # max number of PCs is min(objects, variables)
	if maxPCs < PCs: PCs = maxPCs # change to maxPCs if PCs > maxPCs
	
	Scores = zeros((rows, PCs), float) # all Scores (T)
	Loadings = zeros((PCs, cols), float) # all Loadings (P)
	
	E = mat(X.copy()) #E[0]  (should already be mean centered)


	if E_matrices:
		Error_matrices = zeros((PCs, rows, cols), float) # all Error matrices (E)
	else:
		explained_var = zeros((PCs))
		tot_explained_var = 0
	
		# total object residual variance for PC[0] (calculating from E[0])
		e_tot0 = 0 # for E[0] the total object residual variance is 100%
		for k in range(rows):
			e_k = array(E[k, :])**2
			e_tot0 += sum(e_k)
			
	

	t = get_column_mat(E) # extract a column
	
   
	
	# do iterations (0, PCs)
	for i in range(PCs):
		convergence = False
		ready_for_compare = False
		E_t = transpose(E)
		
		while not convergence:
			eigenval = float(transpose(t)*t)
			p = (E_t*t) / eigenval # ........................................... step 1
			
			_p = float(transpose(p)*p)        
			p = p * _p**(-0.5) # ............................................... step 2
			t = (E*p) / (transpose(p)*p) # ..................................... step 3
			
			
			eigenval_new = float(transpose(t)*t)
			if not ready_for_compare:
				ready_for_compare = True
			else: # ready for convergence check
				if (eigenval_new - eigenval_old) < threshold*eigenval_new: # ... step 4
					convergence = True           
			eigenval_old = eigenval_new;
			
			
		p = transpose(p)
		E = E - (t*p) # ........................................................ step 5
		
		# add Scores and Loadings for PC[i] to the collection of all PCs
		_t = array(t) # NOT optimal
		Scores[:, i] = _t[:,0]; Loadings[i, :] = p[0,:] # co
	 
		if E_matrices:
			# complete error matrix
			# can calculate object residual variance (row-wise) or variable resiudal variance (column-wise)
			# total residual variance can also be calculated
			Error_matrices[i] = E.copy()
		else:
			# total object residual variance for E[i]
			e_tot = 0
		
		for k in range(rows):
			e_ = zeros((cols), float)
			for k_col in range(cols):
				e_[k_col] = E[k, k_col]*E[k, k_col]
			e_tot += sum(e_)
		tot_obj_residual_var =  (e_tot / e_tot0)
		explained_var[i] = 1 - tot_obj_residual_var - tot_explained_var
		tot_explained_var += explained_var[i]
		
		
		

	if E_matrices:
		return Scores, Loadings, Error_matrices
	return Scores, Loadings, explained_var    



def nipals_arr(X, PCs, threshold, E_matrices):
	"""
	
	PCA by NIPALS using numpy array
	
	
	@param X: 2-dimensional matrix of number data. 
	@type X: numpy array
	
	@param PCs: Number of Principal Components.
	@type PCs: int
	
	@param threshold: Convergence check value. For checking on convergence to zero (e.g. 0.000001). 
	@type threshold: float
	
	@param E_matrices: If E-matrices should be retrieved or not. E-matrices (for each PC) or explained_var (explained variance for each PC).
	@type E_matrices: bool
	
	@return: (Scores, Loadings, E)


	"""
	
	(rows, cols) = shape(X)
	maxPCs = min(rows, cols) # max number of PCs is min(objects, variables)
	if maxPCs < PCs: PCs = maxPCs # change to maxPCs if PCs > maxPCs

	Scores = zeros((rows, PCs), float) # all Scores (T)
	Loadings = zeros((PCs, cols), float) # all Loadings (P)
	
	E = X.copy() #E[0]  (should already be mean centered)
	
	if E_matrices:
		Error_matrices = zeros((PCs, rows, cols), float) # all Error matrices (E)
	else:
		explained_var = zeros((PCs))
		tot_explained_var = 0
	
		# total object residual variance for PC[0] (calculating from E[0])
		e_tot0 = 0 # for E[0] the total object residual variance is 100%
		for k in range(rows):
			e_k = E[k, :]**2
			e_tot0 += sum(e_k)      
			
	

	t = get_column(E) # extract a column
	p = zeros((cols), float)
	
	# do iterations (0, PCs)
	for i in range(PCs):
		convergence = False
		ready_for_compare = False
		E_t = transpose(E)
		
		while not convergence:
			_temp = eigenvalue_vec(t)
			p = mat_prod(E_t, t) / _temp # ..................................... step 1
			
			_temp = eigenvalue_vec(p)**(-0.5)
			p = p * _temp # .................................................... step 2
			
			_temp = eigenvalue_vec(p)
			t = mat_prod(E, p) / _temp # ....................................... step 3
			
			
			eigenval_new = eigenvalue_vec(t)
			if not ready_for_compare:
				ready_for_compare = True
			else: # ready for convergence check
				if (eigenval_new - eigenval_old) < threshold*eigenval_new: # ... step 4
					convergence = True           
			eigenval_old = eigenval_new;

		remove_tp_prod(E, t, p) # .............................................. step 5
		
		# add Scores and Loadings for PC[i] to the collection of all PCs
		Scores[:, i] = t; Loadings[i, :] = p
		
		
		if E_matrices:
			# complete error matrix
			# can calculate object residual variance (row-wise) or variable resiudal variance (column-wise)
			# total residual variance can also be calculated
			Error_matrices[i] = E.copy()
		else:
			# total object residual variance for E[i]
			e_tot = 0
		for k in range(rows):
			e_k = E[k, :]**2
			e_tot += sum(e_k)
		tot_obj_residual_var =  (e_tot / e_tot0)
		explained_var[i] = 1 - tot_obj_residual_var - tot_explained_var
		tot_explained_var += explained_var[i]

	if E_matrices:
		return Scores, Loadings, Error_matrices
	else:
		return Scores, Loadings, explained_var 
	
	

def nipals_c(X, PCs, threshold, E_matrices):
	"""  
	
	PCA by NIPALS using python c extension
	
	@param X: 2-dimensional matrix of number data. 
	@type X: numpy array
	
	@param PCs: Number of Principal Components.
	@type PCs: int
	
	@param threshold: Convergence check value. For checking on convergence to zero (e.g. 0.000001). 
	@type threshold: float
	
	@param E_matrices: If E-matrices should be retrieved or not. E-matrices (for each PC) or explained_var (explained variance for each PC).
	@type E_matrices: bool
	
	@return: (Scores, Loadings, E)
	

	"""
	
	if not import_ok:
		raise ImportError("could not import c_nipals python extension")
	else:
		(rows, cols) = shape(X)

	maxPCs = min(rows, cols) # max number of PCs is min(objects, variables)
	if maxPCs < PCs: PCs = maxPCs # change to maxPCs if PCs > maxPCs

	Scores = zeros((rows, PCs), float) # all Scores (T)
	Loadings = zeros((PCs, cols), float) # all Loadings (P)

	E = X.copy() #E[0]  (should already be mean centered)

	if E_matrices:
		Error_matrices = zeros((PCs, rows, cols), float) # all Error matrices (E)
		c_nipals.nipals2(Scores, Loadings, E, Error_matrices, PCs, threshold)
		return Scores, Loadings, Error_matrices
	else:
		explained_var = c_nipals.nipals(Scores, Loadings, E, PCs, threshold)
	return Scores, Loadings, explained_var



################### Principal Component Analysis (using NIPALS) ###################
def PCA_nipals(X, standardize=True, PCs=10, threshold=0.0001, E_matrices=False):
	"""
	
	PCA by NIPALS and get Scores, Loadings, E
	
	@param X: 2-dimensional matrix of number data. 
	@type X: numpy array
	
	@param standardize: Wheter X should be standardized or not.
	@type standardize: bool
	
	@param PCs: Number of Principal Components.
	@type PCs: int
	
	@param threshold: Convergence check value. For checking on convergence to zero (e.g. 0.000001). 
	@type threshold: float
	
	@param E_matrices: If E-matrices should be retrieved or not. E-matrices (for each PC) or explained_var (explained variance for each PC).
	@type E_matrices: bool
	
	@return: nipals_mat(X, PCs, threshold, E_matrices)

	"""

	X = mean_center(X)
		
	if standardize:
		X = standardization(X)
	
	return nipals_mat(X, PCs, threshold, E_matrices) 
	

def PCA_nipals2(X, standardize=True, PCs=10, threshold=0.0001, E_matrices=False):
	"""
	
	PCA by NIPALS and get Scores, Loadings, E
	
	@param X: 2-dimensional matrix of number data. 
	@type X: numpy array
	
	@param standardize: Wheter X should be standardized or not.
	@type standardize: bool
	
	@param PCs: Number of Principal Components.
	@type PCs: int
	
	@param threshold: Convergence check value. For checking on convergence to zero (e.g. 0.000001). 
	@type threshold: float
	
	@param E_matrices: If E-matrices should be retrieved or not. E-matrices (for each PC) or explained_var (explained variance for each PC).
	@type E_matrices: bool
	
	@return: nipals_arr(X, PCs, threshold, E_matrices)

	"""

	X = mean_center(X)
		
	if standardize:
		X = standardization(X)
	
	return nipals_arr(X, PCs, threshold, E_matrices)     


def PCA_nipals_c(X, standardize=True, PCs=10, threshold=0.0001, E_matrices=False):
	"""
	
	PCA by NIPALS and get Scores, Loadings, E
	
	@param X: 2-dimensional matrix of number data. 
	@type X: numpy array
	
	@param standardize: Wheter X should be standardized or not.
	@type standardize: bool
	
	@param PCs: Number of Principal Components.
	@type PCs: int
	
	@param threshold: Convergence check value. For checking on convergence to zero (e.g. 0.000001). 
	@type threshold: float
	
	@param E_matrices: If E-matrices should be retrieved or not. E-matrices (for each PC) or explained_var (explained variance for each PC).
	@type E_matrices: bool
	
	@return: nipals_c(X, PCs, threshold, E_matrices)

	"""

	""" USING C PYTHON EXTENSION """
	X = mean_center(X)
		
	if standardize:
		X = standardization(X)
	
	return nipals_c(X, PCs, threshold, E_matrices)      
	
	
	
################### Principal Component Analysis (using SVD) ###################
def PCA_svd(X, standardize=True):
	"""
	
	PCA by SVD and get Scores, Loadings, E
	
	@param X: 2-dimensional matrix of number data. 
	@type X: numpy array
	
	@param standardize: Wheter X should be standardized or not.
	@type standardize: bool
	
	@return: (Scores, Loadings, explained_var)

	"""

	X = mean_center(X)
	#print X
		
	if standardize:
		X = standardization(X)    
		
	(rows, cols) = shape(X)
	
	# Singular Value Decomposition
	[U, S, V] = svd(X)
	
	# adjust if matrix shape does not match:
	if shape(S)[0] < shape(U)[0]: U = U[:, 0:shape(S)[0]]
	
	Scores = U * S # all Scores (T)
	Loadings = V # all Loadings (P)
	
	variances = S**2 / cols
	variances_sum = sum(variances)
	explained_var = variances / variances_sum
	
	return Scores, Loadings, explained_var



################### Correlation Loadings ###################
def CorrelationLoadings(X, Scores):
	"""
	Get correlation loadings matrix based on Scores (T of PCA) and X (original variables, not mean centered).
	
	
	@param X: 2-dimensional matrix of number data. 
	@type X: numpy array
	
	@param Scores: Scores of PCA (T).
	@type Scores: numpy array
	

	@return: Returns the correlation loadings matrix

	"""
		
	# Creates empty matrix for correlation loadings
	PCs = shape(Scores)[1] # number of PCs
	rows = shape(X)[1] # number of objects (rows) in X
	CorrLoadings = zeros((PCs, rows), float)
		
	# Calculates correlation loadings with formula:
	# correlation = cov(x,y)/(std(x)*std(y))
		
	# For each PC in score matrix
	for i in range(PCs):
		Scores_PC_i = Scores[:, i] # Scores for PC[i]
			
		# For each variable/attribute in X 
		for row in range(rows):
			orig_vars = X[:, row] # column of variables in X
			corrs = corrcoef(Scores_PC_i, orig_vars)
			CorrLoadings[i, row] = corrs[0,1]
				
	return CorrLoadings        



#!/usr/bin/env python3
"""
2012.12.15 data structure for AssociationLocus in PyTablesMatrixFile format
 
"""
import sys, os, math
import csv
import tables
from tables import UInt64Col, StringCol, Float64Col
import numpy
from palos.utils import PassingData, PassingDataList
from palos.ProcessOptions import  ProcessOptions
from palos.io.YHPyTables import YHTable, YHFile, castPyTablesRowIntoPassingData
from palos.algorithm.RBTree import RBDict
from palos.polymorphism.CNV import CNVCompare, CNVSegmentBinarySearchTreeKey, get_overlap_ratio
from . AssociationPeak import AssociationPeakTable

class AssociationLocusTable(tables.IsDescription):
	"""
	2013.1.28 bugfix, was pos=3 for no_of_peaks (same as stop), now change it to pos=4, and increment others accordingly
	"""
	id = UInt64Col(pos=0)
	chromosome = StringCol(64, pos=1)	#64 byte-long
	start = UInt64Col(pos=2)
	stop = UInt64Col(pos=3)
	no_of_peaks = UInt64Col(pos=4)
	connectivity = Float64Col(pos=5)
	no_of_results = UInt64Col(pos=6)
	phenotype_id_ls_in_str = StringCol(1000, pos=7)

class AssociationLocus2PeakTable(tables.IsDescription):
	id = UInt64Col(pos=0)
	association_locus_id = UInt64Col(pos=1)
	association_peak = AssociationPeakTable()


class AssociationLocusTableFile(YHFile):

	"""
	Examples:
		associationLocusTableFile = AssociationLocusTableFile(self.outputFname, mode='w')
		associationLocusTableFile.addAttributeDict(attributeDict)
		associationLocusTableFile.appendAssociationPeak(association_peak_ls=association_peak_ls)
		
		#for read-only
		associationLocusTableFile = AssociationLocusTableFile(path, mode='r')
		rbDict = associationLocusTableFile.associationLocusRBDict
		
		associationLocusTableFile = AssociationLocusTableFile(path, mode='r', constructLocusRBDict=False)	#don't need it
		call_method_id_ls = associationLocusTableFile.getAttribute('call_method_id_ls')
	"""
	def __init__(self, path=None, mode='r', \
				tableName='association_locus', groupNamePrefix='group', tableNamePrefix='table',\
				filters=None, autoRead=True, autoWrite=True, \
				locus2PeakTableName='association_locus2peak', locusPadding=0, constructLocusRBDict=True,\
				**keywords):
		
		self.constructLocusRBDict = constructLocusRBDict
		self.locus2PeakTableName = locus2PeakTableName
		self.locusPadding = locusPadding
		self.associationLocusRBDict = None
		
		YHFile.__init__(self, path=path, mode=mode, \
				tableName=tableName, groupNamePrefix=groupNamePrefix, tableNamePrefix=tableNamePrefix,\
				rowDefinition=None, filters=filters, debug=0, report=0,\
				autoRead=False, autoWrite=False)
		
		#to overwrite self.autoRead that is set by YHFile.__init__
		self.autoRead = autoRead
		self.autoWrite = autoWrite
		
		if self.autoRead and (self.mode=='r' or self.mode=='a'):
			self.associationLocusTable = self.getTableObject(tableName=self.tableName)
			self.associationLocus2PeakTable = self.getTableObject(tableName=self.locus2PeakTableName)
			if self.constructLocusRBDict:
				self.associationLocusRBDict = self._readInData(tableName=self.tableName, tableObject=self.associationLocusTable)
		elif mode == 'w':
			self.associationLocusTable = self.createNewTable(tableName=self.tableName, rowDefinition=AssociationLocusTable,\
													expectedrows=50000)
			self.associationLocus2PeakTable = self.createNewTable(tableName=self.locus2PeakTableName, \
													rowDefinition=AssociationLocus2PeakTable, expectedrows=500000)
	
	def _readInData(self, tableName=None, tableObject=None, bugfixType=None):
		"""
		2013.1.28 added argument bugfixType (default is None)
			1: swap stop & no_of_peaks, an earlier bug exchanged the positions of the two.
		2013.1.26 added phenotype_id_set in the node
		2012.11.25
			similar to constructAssociationPeakRBDictFromHDF5File
		"""
		if tableName is None:
			tableName = self.tableName
		YHFile._readInData(self, tableName=tableName, tableObject=tableObject)
		if not self.constructLocusRBDict:
			return
		
		locusPadding = self.locusPadding
		sys.stderr.write("Constructing association-locus RBDict (locusPadding=%s) ..."%(locusPadding))
		if tableObject is None:
			tableObject = self.getTableObject(tableName=tableName)
		associationLocusRBDict = RBDict()
		associationLocusRBDict.locusPadding = locusPadding
		associationLocusRBDict.HDF5AttributeNameLs = []
		
		for attributeName, value in tableObject.getAttributes().items():
			associationLocusRBDict.HDF5AttributeNameLs.append(attributeName)
			setattr(associationLocusRBDict, attributeName, value)
		
		counter = 0
		real_counter = 0
		for rowPointer in tableObject:
			row = castPyTablesRowIntoPassingData(rowPointer)
			if not row.chromosome:	#empty chromosome, which happens when path contains no valid locus, but the default null locus (only one).
				continue
			counter += 1
			phenotype_id_ls = row.phenotype_id_ls_in_str.split(',')
			phenotype_id_set = set(map(int, phenotype_id_ls))
			if bugfixType==1:
				#2013.1.28 old association-loci file have two columns swapped. run this to correct it.
				# a function in variation/src/misc.py is written:
				#	DB250k.correctAssociationLocusFileFormat(db_250k=db_250k, data_dir=None)
				rowPointer['stop'] = row.no_of_peaks
				rowPointer['no_of_peaks'] = row.stop
				rowPointer.update()
				row.no_of_peaks = rowPointer['no_of_peaks']
				row.stop = rowPointer['stop']
			segmentKey = CNVSegmentBinarySearchTreeKey(chromosome=row.chromosome, \
							span_ls=[max(1, row.start - locusPadding), row.stop + locusPadding], \
							min_reciprocal_overlap=1, no_of_peaks=row.no_of_peaks, \
							no_of_results=row.no_of_results, connectivity=row.connectivity,\
							phenotype_id_set=phenotype_id_set, locus_id=row.id)
							#2010-8-17 overlapping keys are regarded as separate instances as long as they are not identical.
			if segmentKey not in associationLocusRBDict:
				associationLocusRBDict[segmentKey] = []
			associationLocusRBDict[segmentKey].append(row)
		sys.stderr.write("%s peaks in %s spans.\n"%(counter, len(associationLocusRBDict)))
		self.associationLocusRBDict = associationLocusRBDict
		return associationLocusRBDict
	
	def appendAssociationLoci(self, associationLocusList=None):
		"""
		2012.12.10
			for each locus, output the association peaks that fall into the locus.
				for each association peak, include 
					* result-id 
					* phenotype id
					* chromosome
					* start
					* stop
					* start_locus
					* stop_locus
					* no_of_loci
					* peak_locus
					* peak-score
		2012.11.20
		"""
		sys.stderr.write("Saving %s association loci into file %s ..."%(len(associationLocusList), self.path))
		#add neighbor_distance, max_neighbor_distance, min_MAF, min_score, ground_score as attributes
		#2012.11.28 sort it
		associationLocusList.sort()
		for associationLocus in associationLocusList:
			#dataTuple = (associationLocus.chromosome, associationLocus.start, associationLocus.stop, associationLocus.no_of_peaks,\
			#			associationLocus.connectivity, associationLocus.no_of_results, associationLocus.phenotype_id_ls_in_str)
			self.associationLocusTable.writeOneCell(associationLocus, cellType=2)
			associationLocus2PeakTableRow = self.associationLocus2PeakTable.row
			for association_peak in associationLocus.association_peak_ls:
				self.associationLocus2PeakTable.no_of_rows += 1
				associationLocus2PeakTableRow['id'] = self.associationLocus2PeakTable.no_of_rows
				associationLocus2PeakTableRow['association_locus_id'] = self.associationLocusTable.no_of_rows
				associationLocus2PeakTableRow['association_peak/id'] = association_peak['id']
				associationLocus2PeakTableRow['association_peak/chromosome'] = association_peak['chromosome']
				associationLocus2PeakTableRow['association_peak/start'] = association_peak['start']
				associationLocus2PeakTableRow['association_peak/stop'] = association_peak['stop']
				associationLocus2PeakTableRow['association_peak/start_locus_id'] = association_peak['start_locus_id']
				associationLocus2PeakTableRow['association_peak/stop_locus_id'] = association_peak['stop_locus_id']
				associationLocus2PeakTableRow['association_peak/no_of_loci'] = association_peak['no_of_loci']
				associationLocus2PeakTableRow['association_peak/peak_locus_id'] = association_peak['peak_locus_id']
				associationLocus2PeakTableRow['association_peak/peak_score'] = association_peak['peak_score']
				associationLocus2PeakTableRow.append()
				#self.associationLocus2PeakTable.writeOneCell(oneCell, cellType=3)	#doesn't work. nested type shows up as one column in colnames.
		sys.stderr.write("\n")
	
	def fetchOneLocusGivenID(self, locus_id=None):
		"""
		2013.1.27 query the file by the AssociationLocus.id
		"""
		rows = self.readWhere('(id=%s)'%(locus_id))
		if len(rows)>1:
			sys.stderr.write("Warning: %s (>1) rows found with id=%s. Return 1st one.\n"%(len(rows), locus_id))
		elif len(rows)==0:
			#sys.stderr.write("Warning: nothing found with id=%s.\n"%(locus_id))
			return None
		return rows[0]
	
	def fetchOneLocusGivenChrStartStop(self, chromosome=None, start=None, stop=None):
		"""
		2013.1.27
		"""
		rows = self.readWhere('(chromosome=%s) & (start=%s) & (stop=%s)'%(chromosome, start, stop))
		if len(rows)>1:
			sys.stderr.write("Warning: %s (>1) rows found with chr=%s, start=%s, stop=%s. Return 1st one.\n"%\
							(len(rows), chromosome, start, stop))
		elif len(rows)==0:
			#sys.stderr.write("Warning: nothing found with id=%s.\n"%(locus_id))
			return None
		return rows[0]
